# -*- coding: utf-8 -*-

"""
Class for dealing with SRL data.
"""

from collections import defaultdict
#import cPickle
import _pickle
import logging
import re
import os
import numpy as np
#from itertools import izip

from .. import attributes
from .. import utils
from ..word_dictionary import WordDictionary
from .. import reader

class ConllPos(object):
    """
    Dummy class for storing the position of each field in a
    CoNLL data file.
    """
    id = 0
    word = 1
    lemma = 2
    pos = 3
    morph = 4
    parse = 7
    pred = 8
    semantic_role = 9

class SRLReader(reader.TaggerReader):
    
    def __init__(self, md=None, filename=None, only_boundaries=False, 
                 only_classify=False, only_predicates=False):
        """
        The reader will read sentences from a given file. This file must
        be in the correct format (one token per line, columns indicating
        which tokens are predicates and their argument structure).
        
        :param filename: a file with CoNLL-like format data. If it is None,
            the reader will be created with no data.
        :param only_boundaries: train to identify only argument boundaries
        :param only_classify: train to classify pre-determined argument
        :param only_predicates: train to identify only predicates
        """
        
        if only_boundaries:
            self.taskname = 'srl_boundary'
            self._generate_iobes_dictionary()
        elif only_classify:
            self.taskname = 'srl_classify'
        elif only_predicates:
            self.taskname = 'srl_predicates'
            self._generate_predicate_id_dictionary()
        else:
            self.taskname = 'srl'
        
        self.rare_tag = 'O'
        if filename is not None:
            self._read_conll(filename)
            self._clean_text()
        
        super(SRLReader, self).__init__(md)
        
    
    @property
    def task(self):
        """
        Abstract Base Class (ABC) attribute.
        """
        return self.taskname
    
    
    def _read_conll(self, filename):
        '''
        Read a file in CoNLL format and extracts semantic role tags
        for each token.
        '''
        lines = []
        with open(filename, 'rb') as f:
            for line in f:
                line = unicode(line, 'utf-8').strip()
                lines.append(line)
        
        self.sentences = []
        self.predicates = []
        tokens = []
        sent_predicates = []
        sent_tags = []
        token_number = 0
        
        for line in lines:
            line = line.strip()
            
            if line == '':
                # blank line between sentences
                if len(tokens) > 0:
                    sentence = (tokens, sent_tags)
                    self.sentences.append(sentence)
                    self.predicates.append(np.array(sent_predicates))
                    tokens = []
                    sent_predicates = []
                    sent_tags = []
                    token_number = 0
                
                continue
            
            fields = line.split()
            word = fields[ConllPos.word]
            lemma = fields[ConllPos.lemma]
            pos = fields[ConllPos.pos].lower()
            is_predicate = fields[ConllPos.pred] != '-'
            tags = fields[ConllPos.semantic_role:]
            
            # if this is the first token in the sentence, find out how many predicates
            # are there. initialize a list for each of them.
            if sent_tags == []:
                expected_roles = []
                for tag in tags:
                    tag, expected_role = self._read_role(tag, 'O', True)
                    sent_tags.append([tag])
                    expected_roles.append(expected_role)
            else:
                for i, tag in enumerate(tags):
                    expected_role = expected_roles[i]
                    tag, expected_role = self._read_role(tag, expected_role, True)
                    sent_tags[i].append(tag)
                    expected_roles[i] = expected_role
            
            token = attributes.Token(word, lemma, pos)
            tokens.append(token)
            if is_predicate:
                sent_predicates.append(token_number)
            
            token_number += 1
        
        if len(tokens) > 0:
            # last sentence
            sentence = (tokens, sent_tags)
            self.sentences.append(sentence)
            self.predicates.append(np.array(sent_predicates))
    
    @classmethod
    def _read_role(cls, role, expected_role, remove_continuation):
        """
        Reads the next semantic role from a CoNLL-style file.
        
        :param role: what is read from the conll file (something like
            *, (A0* or *)
        :param role: the expected role if a * is found
        :param remove_countinuation: removes the C- from non-continuous
            arguments. C-A0 becomes A0.
        :return a tuple (role, expected next role)
        """
        if role == '*':
            # signals continuation of the last block
            role = expected_role
        elif role == '*)':
            # finishes block
            role = expected_role
            expected_role = 'O'
        else:
            # verifies if it is a single argument
            match = re.search('\(([-\w]+)\*\)', role)
            if match:
                role = match.group(1)
                expected_role = 'O'
            else:
                # verifies if it opens an argument
                match = re.search('\(([-\w]+)\*', role)
                if match:
                    role = match.group(1)
                    expected_role = role
                else:
                    raise ValueError('Unexpected role data: %s' % role)
        
        if role.startswith('C-') and remove_continuation:
            # removes C-
            role = role[2:]
            
        return (role, expected_role)

    def extend(self, data):
        """
        Adds more data to the reader.
        :param data: a list of tuples in the format (tokens, tags, predicates), 
        one for each sentence.
        """
        self.sentences.extend([(sent, tags) for sent, tags, _ in data])
        self.predicates.extend([np.array(preds) for _, _, preds in data])
    
    def load_or_create_tag_dict(self):
        """
        In the case of SRL argument classification or one step SRL, try to 
        load the tag dictionary. If the file with the tags is not present,
        a new one is created from the available sentences. 
        
        In the case of argument detection or predicate detection, 
        this function does nothing.
        """
        if self.task == 'srl_predicates' or self.task == 'srl_boundary':
            return
        
        # only SRL as one step uses IOB tags
        iob = self.task == 'srl'
        if os.path.isfile(self.md.paths['srl_tags']):
            self.load_tag_dict(iob=iob)
            return
        
        self._create_tag_dict(iob)
        logger = logging.getLogger('Logger')
        logger.info('Created SRL tag dictionary')
    
    def _create_tag_dict(self, iob=False):
        """
        Examine the available sentences and create a tag dictionary.
        
        :param iob: If True, this function will generate an entry for B-[tag] 
            and one for I-[tag], except for the tag 'O'.
        """
        logger = logging.getLogger("Logger")
        tags = {tag
                for _, tag_groups in self.sentences
                for tags in tag_groups
                for tag in tags}
        
        # create a dictionary now even if uses IOB, in order to save it in 
        # a deterministic order
        self.tag_dict = {tag: code for code, tag in enumerate(tags)}
        reader.save_tag_dict(self.md.paths['srl_tags'], self.tag_dict)
        logger.debug("Saved SRL tag dictionary.")
        if not iob:
            return
        
        # insert I- and B- preserving the ordering
        new_dict = {}
        code = 0
        for tag in sorted(self.tag_dict, key=self.tag_dict.get):
            if tag == 'O':
                new_dict[tag] = code
            else:
                new_dict['B-%s' % tag] = code
                code += 1
                new_dict['I-%s' % tag] = code
                
            code += 1
        
        self.tag_dict = new_dict
    
    def load_tag_dict(self, filename=None, iob=False):
        """
        Loads the tag dictionary from the default file. The dictionary file should
        have one tag per line. 
        
        :param iob: If True, this function will generate an entry for B-[tag] 
            and one for I-[tag], except for the tag 'O'.
        """
        if self.task == 'srl_predicates' or self.task == 'srl_boundary':
            return
        
        if filename is None:
            filename = self.md.paths['srl_tags']
        
        if not iob:
            super(SRLReader, self).load_tag_dict(filename)
            return
            
        self.tag_dict = {}
        code = 0
        with open(filename, 'rb') as f:
            for tag in f:
                tag = unicode(tag, 'utf-8').strip()
                if tag == '':
                    continue
                
                if tag == 'O':
                    self.tag_dict[tag] = code
                else:
                    self.tag_dict['B-%s' % tag] = code
                    code += 1
                    self.tag_dict['I-%s' % tag] = code
                
                code += 1
        
        if 'O' not in self.tag_dict:
            self.tag_dict['O'] = code
    
    def _generate_iobes_dictionary(self):
        """
        Generate the reader's tag dictionary mapping the IOBES tags to numeric codes.
        """
        self.tag_dict = {tag: code for code, tag in enumerate('IOBES')}
    
    def _generate_predicate_id_dictionary(self):
        """
        Generate a tag dictionary for identifying predicates.
        It has two tags: V for predicates and O for others.
        """
        self.tag_dict = {'O': 0, 'V': 1}
    
    def generate_dictionary(self, dict_size=None, minimum_occurrences=2):
        """
        Generates a token dictionary based on the given sentences.
        
        :param dict_size: Max number of tokens to be included in the dictionary.
        :param minimum_occurrences: Minimum number of times that a token must
            appear in the text in order to be included in the dictionary.
        """
        logger = logging.getLogger("Logger")
        all_tokens = [token.word
                      for tokens, _ in self.sentences
                      for token in tokens]
        self.word_dict = WordDictionary(all_tokens, dict_size, minimum_occurrences)
        logger.info("Created dictionary with %d tokens" % self.word_dict.num_tokens)
    
    def _clean_text(self):
        """
        Cleans the sentences text, replacing numbers for a keyword, different
        kinds of quotation marks for a single one, etc.
        """
        for sent, _ in self.sentences:
            for i, token in enumerate(sent):
                new_word = utils.clean_text(token.word, correct=False)
                new_lemma = utils.clean_text(token.lemma, correct=False) 
                token.word = new_word
                token.lemma = new_lemma
                sent[i] = token

    def create_converter(self):
        """
        This function overrides the TextReader's one in order to deal with Token
        objects instead of raw strings.
        """
        self.converter = attributes.TokenConverter()
        
        if self.md.use_lemma:
            # look up word lemmas 
            word_lookup = lambda t: self.word_dict.get(t.lemma)
        else:
            # look up the word itself
            word_lookup = lambda t: self.word_dict.get(t.word)
             
        self.converter.add_extractor(word_lookup)
        
        if self.md.use_caps:
            caps_lookup = lambda t: attributes.get_capitalization(t.word)
            self.converter.add_extractor(caps_lookup)
        
        if self.md.use_pos:
            with open(self.md.paths['pos_tag_dict']) as f:
                pos_dict = cPickle.load(f)
                
            pos_def_dict = defaultdict(lambda: pos_dict['other'])
            pos_def_dict.update(pos_dict)
            pos_lookup = lambda t: pos_def_dict[t.pos]
            self.converter.add_extractor(pos_lookup)
        
        if self.md.use_chunk:
            with open(self.md.paths['chunk_tag_dict']) as f:
                chunk_dict = cPickle.load(f)
            
            chunk_def_dict = defaultdict(lambda: chunk_dict['O'])
            chunk_def_dict.update(chunk_dict)
            chunk_lookup = lambda t: chunk_def_dict[t.chunk]
            self.converter.add_extractor(chunk_lookup)
    
    def generate_tag_dict(self):
        """
        Generates a tag dictionary that converts the tag itself
        to an index to be used in the neural network.
        """
        self.tagset = set(tag
                          for _, props in self.sentences
                          for prop in props
                          for tag in prop)
        
        self.tag_dict = dict( zip( self.tagset,
                                   xrange(len(self.tagset))
                                   )
                             )
    
    def _remove_tag_names(self):
        """Removes the actual tag names, leaving only IOB or IOBES block delimiters."""
        for _, propositions in self.sentences:
            for tags in propositions:
                for i, tag in enumerate(tags):
                    tags[i] = tag[0]
    
    def _codify_sentences(self):
        """Internal helper function."""
        new_sentences = []
        self.tags = []
        
        for (sent, props), preds in izip(self.sentences, self.predicates):
            new_sent = []
            sentence_tags = []
            
            for token in sent:
                new_token = self.converter.convert(token)
                new_sent.append(new_token)
            
            if self.task == 'srl_predicates':    
                sentence_tags = np.zeros(len(sent), np.int)
                if len(preds) > 0:
                    sentence_tags[preds] = 1
            else:
                for prop in props:
                    # for classifying arguments, leave the names. they will be changed later
                    if self.task == 'srl_classify':
                        prop_tags = prop
                    else:
                        prop_tags = np.array([self.tag_dict[tag] for tag in prop])
                    sentence_tags.append(prop_tags)
            
            new_sentences.append(np.array(new_sent))
            self.tags.append(sentence_tags)
        
        self.sentences = new_sentences
        self.codified = True
    
    def codify_sentences(self):
        """
        Converts each token in each sequence into indices to their feature vectors
        in feature matrices. The previous sentences as text are not accessible anymore.
        Tags are also encoded. This function takes care of the case of classifying 
        pre-delimited arguments.
        """
        if self.converter is None:
            self.create_converter()
        
        self._codify_sentences()
        self.arg_limits = []
        
        if self.task == 'srl_classify':
            # generate the tags for each argument
            start = 0
            end = 0
            
            for i, propositions in enumerate(self.tags):
                new_sent_tags = []
                sent_args = []
                
                for prop_tags in propositions:
                    
                    new_prop_tags = []
                    prop_args = []
                    last_tag = 'O'
                    
                    for j, tag in enumerate(prop_tags):
                        if tag != last_tag:
                            # if we were inside an argument, it ended
                            # we may have started a new
                            if last_tag != 'O':
                                end = j - 1
                                prop_args.append(np.array([start, end]))
                            
                            if tag != 'O':
                                start = j
                                new_prop_tags.append(self.tag_dict[tag])
                            
                        last_tag = tag
                    else:
                        # after last iteration, check the last tag
                        if last_tag != 'O':
                            end = j
                            prop_args.append(np.array([start, end]))
                    
                    sent_args.append(np.array(prop_args))
                    new_sent_tags.append(np.array(new_prop_tags))
                
                self.arg_limits.append(sent_args)
                self.tags[i] = new_sent_tags
                     
    
    def convert_tags(self, scheme, update_tag_dict=True, only_boundaries=False):
        """
        Replaces each word label with an IOB or IOBES version, appending a prefix
        to them. 
        
        :param scheme: IOB or IOBES (In, Other, Begin, End, Single).
        :param update_dict: whether to update or not the tag dictionary after
            converting the tags.
        :param only_boundaries: if True, only leaves the IOBES tags and remove
            the actual tags. Also, avoid updating the tag dict.
        """
        scheme = scheme.lower()
        if scheme not in ('iob', 'iobes'):
            raise ValueError("Unknown tagging scheme: %s" % scheme)
        
        for _, props in self.sentences:
            for prop in props:
                
                last_tag = None
                for i, tag in enumerate(prop):
                    
                    if tag == 'O':
                        # O tag is independent from IBES
                        last_tag = tag 
                        continue
                    
                    try:
                        next_tag = prop[i + 1]
                    except IndexError:
                        # last word already
                        next_tag = None
                     
                    if tag != last_tag:
                        # a new block starts here. 
                        last_tag = tag
                        if scheme == 'iob' or next_tag == tag:
                            prop[i] = 'B-%s' % tag
                        else:
                            prop[i] = 'S-%s' % tag
                    else:
                        # the block continues. 
                        if scheme == 'iob' or next_tag == tag:
                            prop[i] = 'I-%s' % tag
                        else:
                            prop[i] = 'E-%s' % tag
            
        if only_boundaries:
            self._remove_tag_names()
        elif update_tag_dict:
            self.generate_tag_dict()
        else:
            # treat any tag not appearing in the tag dictionary as O
            actual_tagset = {tag for _, props in self.sentences for prop in props for tag in prop}
            for tag in actual_tagset:
                if tag not in self.tag_dict:
                    self.tag_dict[tag] = self.tag_dict[self.rare_tag]
    



import requests
import json
from datetime import datetime
import hmac
import base64
import uuid
import time
import re
from bs4 import BeautifulSoup as bs


class papago: 

	def __init__(self):
		
		
		response=requests.get('https://papago.naver.com')
		html=bs(response.text,'html.parser')
		pattern1=r'/vendors~main.*chunk.js'

		for tmp in html.find_all('script'):
			tmp=str(tmp)
			m=re.search(pattern1,tmp)
			if m is not None:
				a=m.group()

		js_url='https://papago.naver.com'+str(a)
		rest=requests.get(js_url)
		org=rest.text
		pattern2=r'AUTH_KEY:[\s]*"[\w.]+"'
		self.match=str(re.findall(pattern2,org)).split('"')[1]

	# headers 보안키 생성
	def hmac_md5(self,key, s):
		return base64.b64encode(hmac.new(key.encode('utf-8'), s.encode('utf-8'), 'MD5').digest()).decode()


	def translate(self,data,source,target):

		url = 'https://papago.naver.com/apis/n2mt/translate'
		AUTH_KEY = self.match

		dt = datetime.now()
		timestamp = str(round(dt.timestamp()*1000))

		# 고정 값을 사용할 시 서버로 부터 차단을 방지
		deviceId = str(uuid.uuid4())

		headers = {
				'authorization': 'PPG ' + deviceId + ':' + self.hmac_md5(AUTH_KEY, deviceId + '\n' + url + '\n' + timestamp),
				'timestamp': timestamp
				}

		form_data = {
				'deviceId': deviceId,
				'locale': 'ko',
				'dict': 'true',
				'dictDisplay': 30,
				'honorific': 'false',
				'instant': 'false',
				'paging': 'false',
				'source': source,
				'target': target,
				'text': data
				}

		res_data = requests.post(url, data=form_data, headers=headers)

		#papago 번역 결과물 전체 확인
		#print("\n\n\n",res_data.json())

		return res_data.json()['translatedText']


	def e2k(self,sent_list):

		patient = 0
		return_list=[]

		for line in sent_list:
			line = line.strip()
			try:
				text = self.translate(line,'en','ko') ## translatin
			except (KeyError,requests.exceptions.ConnectionError) as e:
				if patient > 5:
					ofp.close()
					exit() ## Error가 5번 이상 누적되면 종료
				patient += 1
				time.sleep(30) ## 에러 발생 시 1시간 대기
				continue

			return_list.append(text)
		
		#print(json.dumps(result, ensure_ascii=False), flush=True, file=ofp) ## json line 형식으로 저장
		return return_list


	def k2e(self,sent_list):

		patient = 0
		return_list=[]

		for line in sent_list:
			line = line.strip()
			try:
				text = self.translate(line,'ko','en') ## translatin
			except (KeyError,requests.exceptions.ConnectionError) as e:
				if patient > 5:
					ofp.close()
					exit() ## Error가 5번 이상 누적되면 종료
				patient += 1
				time.sleep(30) ## 에러 발생 시 1시간 대기
				continue

			return_list.append(text)
		
		#print(json.dumps(result, ensure_ascii=False), flush=True, file=ofp) ## json line 형식으로 저장
		return return_list






# -*- coding: utf-8 -*-

"""
A convolutional neural network for NLP tagging tasks like SRL.
It employs feature tables to store feature vectors for each token.
"""

import numpy as np
cimport numpy as np

cdef class ConvolutionalNetwork(Network):
    
    # transition and distance feature tables
    cdef public np.ndarray target_dist_table, pred_dist_table
    cdef readonly np.ndarray target_dist_weights, pred_dist_weights
    cdef readonly int target_dist_offset, pred_dist_offset
    cdef readonly np.ndarray target_dist_lookup, pred_dist_lookup
    cdef readonly np.ndarray target_convolution_lookup, pred_convolution_lookup
    cdef readonly np.ndarray target_dist_deltas, pred_dist_deltas
    
    # the second hidden layer
    cdef readonly int hidden2_size
    cdef readonly np.ndarray hidden2_weights, hidden2_bias
    cdef readonly np.ndarray hidden2_values
    cdef readonly np.ndarray hidden2_before_activation, hidden_before_activation
    
    # lookup of convolution values (the same for each sentence, used to save time)
    cdef np.ndarray convolution_lookup
        
    # maximum convolution indices
    cdef readonly np.ndarray max_indices
    
    # number of targets (all tokens in a sentence or the provided arguments)
    # and variables for argument classifying
    cdef int num_targets
    cdef bool only_classify
    
    # for faster access 
    cdef int half_window
    
    # the convolution gradients 
    cdef np.ndarray hidden_gradients, hidden2_gradients
    cdef np.ndarray input_deltas
    
    # keeping statistics
    cdef int num_sentences
    
    # validation
    cdef list validation_predicates, validation_arguments
    
    @classmethod
    def create_new(cls, feature_tables, target_dist_table, pred_dist_table, 
                   int word_window, int hidden1_size, int hidden2_size, int output_size):
        """Creates a new convolutional neural network."""
        # sum the number of features in all tables except for distance 
        cdef int input_size = sum(table.shape[1] for table in feature_tables)
        input_size *= word_window
        
        dist_features_per_token = target_dist_table.shape[1] + pred_dist_table.shape[1]
        input_size_with_distance = input_size + (word_window * dist_features_per_token)
        
        # creates the weight matrices
        high = 2.38 / np.sqrt(input_size_with_distance) # [Bottou-88]
        hidden_weights = np.random.uniform(-high, high, (hidden1_size, input_size))
        
        num_dist_features = word_window * target_dist_table.shape[1]
        target_dist_weights = np.random.uniform(-high, high, (num_dist_features, hidden1_size))
        num_dist_features = word_window * pred_dist_table.shape[1]
        pred_dist_weights = np.random.uniform(-high, high, (num_dist_features, hidden1_size))
        
        high = 2.38 / np.sqrt(hidden1_size)
        hidden_bias = np.random.uniform(-high, high, hidden1_size)
        
        if hidden2_size > 0:
            hidden2_weights = np.random.uniform(-high, high, (hidden2_size, hidden1_size))
            high = 2.38 / np.sqrt(hidden2_size)
            hidden2_bias = np.random.uniform(-high, high, hidden2_size)
            output_dim = (output_size, hidden2_size)
        else:
            hidden2_weights = None
            hidden2_bias = None
            output_dim = (output_size, hidden1_size)
        
        high = 2.38 / np.sqrt(output_dim[1])
        output_weights = np.random.uniform(-high, high, output_dim)
        high = 2.38 / np.sqrt(output_size)
        output_bias = np.random.uniform(-high, high, output_size)
        
        net = cls(word_window, input_size, hidden1_size, hidden2_size, 
                  output_size, hidden_weights, hidden_bias, 
                  target_dist_weights, pred_dist_weights,
                  hidden2_weights, hidden2_bias, 
                  output_weights, output_bias)
        net.feature_tables = feature_tables
        net.target_dist_table = target_dist_table
        net.pred_dist_table = pred_dist_table
        
        return net
    
    def description(self):
        """Returns a textual description of the network."""
        hidden2_size = 0 if self.hidden2_weights is None else self.hidden2_size
        table_dims = [str(t.shape[1]) for t in self.feature_tables]
        table_dims =  ', '.join(table_dims)
        
        dist_table_dims = '%d, %d' % (self.target_dist_table.shape[1], self.pred_dist_table.shape[1])
        
        desc = """
Word window size: %d
Feature table sizes: %s
Distance table sizes (target and predicate): %s
Input layer size: %d
Convolution layer size: %d 
Second hidden layer size: %d
Output size: %d
""" % (self.word_window_size, table_dims, dist_table_dims, self.input_size, self.hidden_size,
       hidden2_size, self.output_size)
        
        return desc
    
    
    def __init__(self, word_window, input_size, hidden1_size, hidden2_size,
                 output_size, hidden1_weights, hidden1_bias, target_dist_weights, 
                 pred_dist_weights, hidden2_weights, hidden2_bias, 
                 output_weights, output_bias):
        super(ConvolutionalNetwork, self).__init__(word_window, input_size, 
                                                   hidden1_size, output_size, 
                                                   hidden1_weights, hidden1_bias, 
                                                   output_weights, output_bias)
        self.half_window = word_window / 2
        self.features_per_token = self.input_size / word_window
        
        self.transitions = None
        self.target_dist_lookup = None
        self.pred_dist_lookup = None
        self.target_dist_weights = target_dist_weights
        self.pred_dist_weights = pred_dist_weights
        
        self.hidden2_size = hidden2_size
        self.hidden2_weights = hidden2_weights
        self.hidden2_bias = hidden2_bias
        
        self.validation_predicates = None
        self.validation_arguments = None
        
        self.use_learning_rate_decay = False
    
    def _generate_save_dict(self):
        """
        Generates a dictionary with all parameters saved by the model.
        It is directly used by the numpy savez function.
        """
        d = dict(hidden_weights=self.hidden_weights,
                 target_dist_table=self.target_dist_table,
                 pred_dist_table=self.pred_dist_table,
                 target_dist_weights=self.target_dist_weights,
                 pred_dist_weights=self.pred_dist_weights,
                 output_weights=self.output_weights,
                 transitions=self.transitions,
                 hidden_bias=self.hidden_bias, output_bias=self.output_bias,
                 word_window_size=self.word_window_size, 
                 input_size=self.input_size, hidden_size=self.hidden_size,
                 output_size=self.output_size, hidden2_size=self.hidden2_size,
                 hidden2_weights=self.hidden2_weights, hidden2_bias=self.hidden2_bias,
                 padding_left=self.padding_left, padding_right=self.padding_right,
                 feature_tables=self.feature_tables)
        return d
    
    def save(self):
        """
        Saves the neural network to a file.
        It will save the weights, biases, sizes, padding and 
        distance tables, and other feature tables.
        """
        data = self._generate_save_dict()
        np.savez(self.network_filename, **data)

    @classmethod
    def _load_from_file(cls, data, filename):
        """
        Internal method for setting data read from a npz file.
        """
        # cython classes don't have the __dict__ attribute
        # so we can't do an elegant self.__dict__.update(data)
        hidden_weights = data['hidden_weights']
        hidden_bias = data['hidden_bias']
        hidden2_weights = data['hidden2_weights']
        
        hidden2_bias = data['hidden2_bias']
        output_weights = data['output_weights']
        output_bias = data['output_bias']
        
        word_window_size = data['word_window_size']
        input_size = data['input_size']
        hidden_size = data['hidden_size']
        hidden2_size = data['hidden2_size']
        output_size = data['output_size']
        
        # numpy stores None as an array containing None and with empty shape
        if hidden2_weights.shape == (): 
            hidden2_weights = None
            hidden2_size = 0
            hidden2_bias = None
        
        nn = cls(word_window_size, input_size, hidden_size, hidden2_size, 
                 output_size, hidden_weights, hidden_bias, 
                 data['target_dist_weights'], data['pred_dist_weights'],
                 hidden2_weights, hidden2_bias, 
                 output_weights, output_bias)
        
        nn.target_dist_table = data['target_dist_table']
        nn.pred_dist_table = data['pred_dist_table']
        #transitions = data['transitions']
        #nn.transitions = transitions if transitions.shape != () else None
        nn.padding_left = data['padding_left']
        nn.padding_right = data['padding_right']
        nn.pre_padding = np.array(int(nn.word_window_size // 2) * [nn.padding_left])
        nn.pos_padding = np.array(int(nn.word_window_size // 2) * [nn.padding_right])
        nn.feature_tables = list(data['feature_tables'])
        nn.network_filename = filename
        
        return nn

    @classmethod
    def load_from_file(cls, filename):
        """
        Loads the neural network from a file.
        It will load weights, biases, sizes, padding and 
        distance tables, and other feature tables.
        """
        data = np.load(filename, allow_pickle=True) # 수정 
        return cls._load_from_file(data, filename)
    
    def _load_parameters(self):
        """
        Loads weights, feature tables, distance tables and 
        transition tables previously saved.
        """
        data = np.load(self.network_filename)
        self.hidden_weights = data['hidden_weights']
        self.hidden_bias = data['hidden_bias']
        self.output_weights = data['output_weights']
        self.output_bias = data['output_bias']
        self.feature_tables = list(data['feature_tables'])
        self.target_dist_table = data['target_dist_table']
        self.pred_dist_table = data['pred_dist_table']
        
        # check if transitions isn't None (numpy saves everything as an array)
        if data['transitions'].shape != ():
            self.transitions = data['transitions']
        else:
            self.transitions = None
            
        # same for second hidden layer weights
        if data['hidden2_weights'].shape != ():
            self.hidden2_weights = data['hidden2_weights']
            self.hidden2_bias = data['hidden2_bias']
        else:
            self.hidden2_weights = None
        
    def set_validation_data(self, list validation_sentences,
                            list validation_predicates,
                            list validation_tags,
                            list validation_arguments=None):
        """
        Sets the data to be used in validation during training. If this function
        is not called before training, the training data itself is used to 
        measure the model's performance.
        """
        self.validation_sentences = validation_sentences
        self.validation_predicates = validation_predicates
        self.validation_tags = validation_tags
        self.validation_arguments = validation_arguments
    
    def train(self, list sentences, list predicates, list tags,  
              int epochs, int epochs_between_reports=0,
              float desired_accuracy=0, list arguments=None):
        """
        Trains the convolutional network. Refer to the basic Network
        train method for detailed explanation.
        
        :param predicates: a list of 1-dim numpy array
            indicating the indices of predicates in each sentence.
        :param arguments: (only for argument classifying) a list of 2-dim
            numpy arrays indicating the start and end of each argument. 
        """
        self.num_sentences = len(sentences)
        self.num_tokens = sum(len(sent) for sent in sentences)
        self.only_classify = arguments is not None
        
        logger = logging.getLogger("Logger")
        logger.info("Training for up to %d epochs" % epochs)
        last_accuracy = 0
        top_accuracy = 0
        last_error = np.Infinity
        
        if self.validation_sentences is None:
            self.set_validation_data(sentences, predicates, tags, arguments)
        
        for i in xrange(epochs):
            self.decrease_learning_rates(i)
            self._train_epoch(sentences, predicates, tags, arguments)
            self._validate()
            
            # Attardi: save model
            if self.accuracy > top_accuracy:
                top_accuracy = self.accuracy
                self.save()
                logger.debug("Saved model")
            elif self.use_learning_rate_decay:
                # this iteration didn't bring improvements; load the last saved model
                # before continuing training with a lower rate
                self._load_parameters()
            
            if (epochs_between_reports > 0 and i % epochs_between_reports == 0) \
                or self.accuracy >= desired_accuracy > 0 \
                or (self.accuracy < last_accuracy and self.error > last_error):
                
                self._print_epoch_report(i + 1)

                if self.accuracy >= desired_accuracy > 0\
                        or (self.accuracy < last_accuracy and self.error > last_error):
                    # accuracy is falling, the network is probably diverging
                    # or overfitting
                    break
            
            last_accuracy = self.accuracy
            last_error = self.error
        
        self.num_sentences = 0
        self.num_tokens = 0
        self._reset_counters()
    
    def _reset_counters(self):
        """
        Reset the performance statistics counters. They are updated during
        each epoch. 
        """
        self.error = 0
        self.skips = 0
        self.float_errors = 0
    
    def _shuffle_data(self, sentences, predicates, tags, arguments=None):
        """
        Shuffle the given training data in place.
        """
        # get the random number generator state in order to shuffle
        # sentences and their tags in the same order
        random_state = np.random.get_state()
        np.random.shuffle(sentences)
        np.random.set_state(random_state)
        np.random.shuffle(predicates)
        np.random.set_state(random_state)
        np.random.shuffle(tags)
        if arguments is not None:
            np.random.set_state(random_state)
            np.random.shuffle(arguments)
        
        
    def _train_epoch(self, sentences, predicates, tags, arguments):
        """Trains for one epoch with all examples."""
        
        self._reset_counters()
        self._shuffle_data(sentences, predicates, tags, arguments)
        if arguments is not None:
            i_args = iter(arguments)
        else:
            sent_args = None
        
        for sent, sent_preds, sent_tags in zip(sentences, predicates, tags):
            if arguments is not None:
                sent_args = next(i_args)
                #sent_args = i_args.next()
            
            try:
                self._tag_sentence(sent, sent_preds, sent_tags, sent_args)
            except FloatingPointError:
                # just ignore the sentence in case of an overflow
                self.float_errors += 1
    
    def tag_sentence(self, np.ndarray sentence, np.ndarray predicates, 
                     list arguments=None, bool logprob=False,
                     bool allow_repeats=True):
        """
        Runs the network for each element in the sentence and returns 
        the sequence of tags.
        
        :param sentence: a 2-dim numpy array, where each item encodes a token.
        :param predicates: a 1-dim numpy array, indicating the position
            of the predicates in the sentence
        :param logprob: a boolean indicating whether to return the 
            log-probability for each answer or not.
        :param allow_repeats: a boolean indicating whether to allow repeated
            argument classes (only for separate argument classification).
        """
        self.only_classify = arguments is not None
        return self._tag_sentence(sentence, predicates, argument_blocks=arguments, 
                                  logprob=logprob, allow_repeats=allow_repeats)
    
    cdef np.ndarray argument_distances(self, positions, argument):
        """
        Calculates the distance from each token in the sentence to the argument.
        """
        distances = positions.copy()
        
        # the ones before the argument
        #print(argument)
        lo = np.less(positions, argument[0])
        distances[lo] -= argument[0]
        
        # the ones after the argument
        hi = np.greater(positions, argument[1])
        distances[hi] -= argument[1]
        
        # the ones inside the argument
        distances[np.logical_not(hi | lo)] = 0
        
        return distances
    
    
    @cython.boundscheck(False)
    @cython.wraparound(False)
    def _sentence_convolution(self, sentence, predicate, argument_blocks=None, 
                              training=False):
        """
        Perform the convolution for a given predicate.
        
        :param sentence: a sequence of tokens, each represented as an array of 
            indices
        :param predicate: the index of the predicate in the sentence
        :param argument_blocks: (used only in SRL argument classification) the
            starting and end positions of all delimited arguments
        :return: the scores for all tokens with respect to the given predicate
        """        
        # store the values found by each convolution neuron here and then find the max
        cdef np.ndarray[FLOAT_t, ndim=2] convolution_values
        
        # a priori scores for all tokens
        cdef np.ndarray[FLOAT_t, ndim=2] scores
        
        # intermediate storage
        cdef np.ndarray[FLOAT_t, ndim=2] input_and_pred_dist_values
        
        self.num_targets = len(sentence) if argument_blocks is None else len(argument_blocks)
        
        # maximum values found by convolution
        self.hidden_values = np.zeros((self.num_targets, self.hidden_size))
        
        if training:
            # hidden sent values: results after tanh
            self.hidden_values = np.zeros((self.num_targets, self.hidden_size))
            self.max_indices = np.empty((self.num_targets, self.hidden_size), int)
            #self.max_indices = np.empty((self.num_targets, self.hidden_size), np.int)
    
        # predicate distances are the same across all targets
        pred_dist_indices = np.arange(len(sentence)) - predicate
        pred_dist_values = self.pred_convolution_lookup.take(pred_dist_indices + self.pred_dist_offset,
                                                             0, mode='clip')
        
        input_and_pred_dist_values = pred_dist_values + self.convolution_lookup
        
        for target in range(self.num_targets):
            # loop over targets and add the weighted distance features to each token
            # this is necessary for the convolution layer
            
            # distance features for each window
            # if we are classifying all tokens, pick the distance to the target
            # if we are classifying arguments, pick the distance to the closest boundary 
            # of the argument (beginning or end)
            if argument_blocks is None:
                target_dist_indices = np.arange(len(sentence)) - target
            else:
                argument = argument_blocks[target]
                #print(argument_blocks)
                target_dist_indices = self.argument_distances(np.arange(len(sentence)), argument)
            
            target_dist_values = self.target_convolution_lookup.take(target_dist_indices + self.target_dist_offset,
                                                                     0, mode='clip')

            convolution_values = target_dist_values + input_and_pred_dist_values
            
            # now, find the maximum values
            if training:
                self.max_indices[target] = convolution_values.argmax(0)
            self.hidden_values[target] = convolution_values.max(0)
            
        # apply the bias and proceed to the next layer
        self.hidden_values += self.hidden_bias
    
        if self.hidden2_weights is not None:
            self.hidden2_values = self.hidden_values.dot(self.hidden2_weights.T) + self.hidden2_bias
        
            if training:
                self.hidden2_before_activation = self.hidden2_values.copy()
    
            hardtanh(self.hidden2_values, inplace=True)
        else:
            # apply non-linearity here
            if training:
                self.hidden_before_activation = self.hidden_values.copy()
            
            self.hidden2_values = self.hidden_values
            hardtanh(self.hidden_values, inplace=True)
            
        scores = self.hidden2_values.dot(self.output_weights.T) + self.output_bias
        
        return scores
    
    def _pre_tagging_setup(self, np.ndarray sentence, bool training):
        """
        Perform some initialization actions before the actual tagging.
        """
        if training:
            # this table will store the values of the neurons for each input token
            # they will be needed during weight adjustments
            self.input_sent_values = np.empty((len(sentence), self.input_size))
        
        # store the convolution values to save time
        self._create_convolution_lookup(sentence, training)
        
        if self.target_dist_lookup is None: self._create_target_lookup()
        if self.pred_dist_lookup is None: self._create_pred_lookup()
        
    
    @cython.boundscheck(False)
    @cython.wraparound(False)
    def _tag_sentence(self, np.ndarray sentence, np.ndarray predicates, 
                      list tags=None, list argument_blocks=None, 
                      bool allow_repeats=True, bool logprob=False):
        """
        Runs the network for every predicate in the sentence.
        Refer to the Network class for more information.
        
        :param tags: this is a list rather than a numpy array because in
            argument classification, each predicate may have a differente number
            of arguments.
        :param argument_blocks: (used only in SRL argument classification) a list
            with the starting and end positions of all delimited arguments (one for 
            each predicate)
        :param predicates: a numpy array with the indices of the predicates in the sentence.
        """
        answer = []
        training = tags is not None
        self._pre_tagging_setup(sentence, training)
        cdef np.ndarray[FLOAT_t, ndim=2] token_scores
        
        for i, predicate in enumerate(predicates):
            pred_arguments = None if not self.only_classify else argument_blocks[i]
            pred_arguments = argument_blocks[i]
            
            token_scores = self._sentence_convolution(sentence, predicate, pred_arguments, training)
            pred_answer = self._viterbi(token_scores, allow_repeats)
        
            if training:
                pred_tags = tags[i]
                if self._calculate_gradients(pred_tags, token_scores):
                    self._backpropagate()
                    self._calculate_input_deltas(sentence, predicate, pred_arguments)
                    self._adjust_weights(predicate, pred_arguments)
                    self._adjust_features(sentence, predicate)
            
            if logprob:
                if self.only_classify:
                    raise NotImplementedError('Confidence measure not implemented for argument classifying')
                
                all_scores = self._calculate_all_scores(token_scores)
                last_token = len(sentence) - 1
                logadd = np.log(np.sum(np.exp(all_scores[last_token])))
                confidence = self.answer_score - logadd
                pred_answer = (pred_answer, confidence)
            
            answer.append(pred_answer)
        
        return answer
        
    def _validate(self):
        """
        Evaluates the network performance, updating its hits count.
        """
        # call it "item" instead of token because the same token may be counted
        # more than once (sentences with multiple predicates)
        num_items = 0
        hits = 0
        
        if self.validation_arguments is not None:
            i_args = iter(self.validation_arguments)
        else:
            sent_args = None
        
        for sent, sent_preds, sent_tags in zip(self.validation_sentences, 
                                                self.validation_predicates, 
                                                self.validation_tags):
            if self.validation_arguments is not None:
                sent_args = next(i_args)
                #sent_args = i_args.next()
            
            answer = self._tag_sentence(sent, sent_preds, None, sent_args)
            for predicate_answer, predicate_tags in zip(answer, sent_tags):
                for net_tag, gold_tag in zip(predicate_answer, predicate_tags):
                    if net_tag == gold_tag:
                        hits += 1
                
                num_items += len(predicate_answer)
        
        self.accuracy = float(hits) / num_items
        # normalize error
        self.error /= num_items
    
    def _calculate_gradients(self, tags, scores):
        """Delegates the call to the appropriate function."""
        if self.only_classify:
            return self._calculate_gradients_classify(tags, scores)
        else:
            return self._calculate_gradients_sll(tags, scores)
    
    def _calculate_gradients_classify(self, tags, scores):
        """
        Calculates the output deltas for each target in a network that only 
        classifies predelimited arguments.
        The aim is to minimize the cost, for each argument:
        logadd(score for all possible tags) - score(correct tag)
        
        :returns: whether a correction is necessary or not.
        """
        self.net_gradients = np.zeros_like(scores, float)
        #self.net_gradients = np.zeros_like(scores, np.float)
        correction = False
        
        for i, tag_scores in enumerate(scores):
            tag = tags[i]
            
            exponentials = np.exp(tag_scores)
            exp_sum = np.sum(exponentials)
            logadd = np.log(exp_sum)
            
            # update the total error 
            error = logadd - tag_scores[tag]
            self.error += error
            
            # like the non-convolutional network, don't adjust weights if the error
            # is too low. An error of 0.01 means a log-prob of -0.01 for the right
            # tag, i.e., more than 99% probability
            if error <= 0.01:
                self.skips += 1
                continue
            
            correction = True
            self.net_gradients[i] = - exponentials / exp_sum
            self.net_gradients[i, tag] += 1
    
        return correction

    def _backpropagate(self):
        """Backpropagates the error gradient."""
        
        # this function only determines the gradients at each layer, without 
        # adjusting weights. This is done because the input features must 
        # be adjusted with the first weight matrix unchanged. 
        
        # gradient[i][j] has the gradient for token i at neuron j
        
        # derivative with respect to the non-linearity layer (tanh)
        dCd_tanh = self.net_gradients.dot(self.output_weights)
        
        if self.hidden2_weights is not None:
            # derivative with respect to the second hidden layer
            dCd_hidden2 = dCd_tanh * hardtanhd(self.hidden2_before_activation)
            self.hidden2_gradients = dCd_hidden2
                        
            self.hidden_gradients = self.hidden2_gradients.dot(self.hidden2_weights)
        else:
            # the non-linearity appears right after the convolution max
            self.hidden_gradients = dCd_tanh * hardtanhd(self.hidden_before_activation)
    
    @cython.boundscheck(False)
    @cython.wraparound(False)
    def _adjust_weights(self, predicate, arguments=None):
        """
        Adjusts the network weights after gradients have been calculated.
        """
        cdef int i
        cdef np.ndarray[FLOAT_t, ndim=1] gradients_t
        cdef np.ndarray[FLOAT_t, ndim=2] last_values, deltas, grad_matrix, input_values
        
        last_values = self.hidden2_values if self.hidden2_weights is not None else self.hidden_values
        deltas = self.net_gradients.T.dot(last_values) * self.learning_rate
        self.output_weights += deltas
        self.output_bias += self.net_gradients.sum(0) * self.learning_rate
        
        if self.hidden2_weights is not None:
            deltas = self.hidden2_gradients.T.dot(self.hidden_values) * self.learning_rate
            self.hidden2_weights += deltas
            self.hidden2_bias += self.hidden2_gradients.sum(0) * self.learning_rate
        
        # now adjust weights from input to convolution. these will be trickier.
        # we need to know which input value to use in the delta formula
        
        # I tried vectorizing this loop but it got a bit slower, probably because
        # of the overhead in building matrices/tensors with the max indices
        for i, neuron_maxes in enumerate(self.max_indices):
            # i indicates the i-th target
              
            gradients_t = self.hidden_gradients[i] * self.learning_rate
            
            # table containing in each line the input values selected for each convolution neuron
            input_values = self.input_sent_values.take(neuron_maxes, 0)
            
            # stack the gradients to multiply all weights for a neuron
            grad_matrix = np.tile(gradients_t, [self.input_size, 1]).T
            self.hidden_weights += grad_matrix * input_values
            
            # target distance weights
            # get the relative distance from each max token to its target
            if arguments is None:
                target_dists = neuron_maxes - i
            else:
                argument = arguments[i]
                target_dists = self.argument_distances(neuron_maxes, argument)
            
            dist_features = self.target_dist_lookup.take(target_dists + self.target_dist_offset, 
                                                         0, mode='clip')
            grad_matrix = np.tile(gradients_t, [self.target_dist_weights.shape[0], 1]).T
            self.target_dist_weights += (grad_matrix * dist_features).T
            
            # predicate distance weights
            # get the distance from each max token to its predicate
            pred_dists = neuron_maxes - predicate
            dist_features = self.pred_dist_lookup.take(pred_dists + self.pred_dist_offset,
                                                       0, mode='clip')
            # try to recycle the grad_matrix if sizes match
            if self.target_dist_weights.shape[0] != self.pred_dist_weights.shape[0]: 
                grad_matrix = np.tile(gradients_t, [self.pred_dist_weights.shape[0], 1]).T
            
            self.pred_dist_weights += (grad_matrix * dist_features).T
        
        self.hidden_bias += self.hidden_gradients.sum(0) * self.learning_rate
        
        # Adjusts the transition scores table with the calculated gradients.
        if not self.only_classify and self.transitions is not None:
            self.transitions += self.trans_gradients * self.learning_rate_trans
            
    @cython.boundscheck(False)
    @cython.wraparound(False)
    def _calculate_input_deltas(self, np.ndarray sentence, int predicate, 
                                 object arguments=None):
        """
        Calculates the input deltas to be applied in the feature tables.
        """
        cdef np.ndarray[FLOAT_t, ndim=2] hidden_gradients, input_gradients
        cdef np.ndarray[FLOAT_t, ndim=2] target_dist_gradients, pred_dist_gradients
        cdef np.ndarray[FLOAT_t, ndim=1] gradients
        cdef np.ndarray[INT_t, ndim=1] convolution_max, target_dists
        
        # matrices accumulating gradients over each target
        # each matrix has a whole window in each line
        input_gradients = np.zeros((len(sentence), self.hidden_size))
        target_dist_gradients = np.zeros((self.target_dist_lookup.shape[0], self.hidden_size))
        pred_dist_gradients = np.zeros((self.pred_dist_lookup.shape[0], self.hidden_size))
        
        # avoid multiplying by the learning rate multiple times
        hidden_gradients = self.hidden_gradients * self.learning_rate_features
        cdef np.ndarray[INT_t, ndim=1] column_numbers = np.arange(self.hidden_size)
        
        for target in range(self.num_targets):
            
            # array with the tokens that yielded the maximum value in each neuron
            # for this target
            convolution_max = self.max_indices[target]
            
            if not self.only_classify:
                target_dists = convolution_max - target
            else:
                argument = arguments[target]
                target_dists = self.argument_distances(convolution_max, argument)
            
            target_dists = np.clip(target_dists + self.target_dist_offset, 0,
                                   self.target_dist_lookup.shape[0] - 1)
            pred_dists = convolution_max - predicate
            pred_dists = np.clip(pred_dists + self.pred_dist_offset, 0,
                                 self.pred_dist_lookup.shape[0] - 1)
            
            gradients = hidden_gradients[target]
            
            # sparse matrix with gradients to be applied over the input
            # line i has the gradients for the i-th token in the sentence
            input_gradients[convolution_max, np.arange(self.hidden_size)] += gradients
            
            # distance deltas
            target_dist_gradients[target_dists, np.arange(self.hidden_size)] += gradients
            pred_dist_gradients[pred_dists, np.arange(self.hidden_size)] += gradients
        
        self.input_deltas = input_gradients.dot(self.hidden_weights)
        self.target_dist_deltas = target_dist_gradients.dot(self.target_dist_weights.T)
        self.pred_dist_deltas = pred_dist_gradients.dot(self.pred_dist_weights.T)
            
        
    def _adjust_features(self, sentence, predicate):
        """Adjusts the features in all feature tables."""
        # compute each token in the window separately and
        # separate the feature deltas into tables
        start_from = 0
        dist_target_from = 0
        dist_pred_from = 0
        
        # number of times that the minimum and maximum distances are repeated
        # in the lookup distance tables
        pre_dist = self.word_window_size
        pos_dist = 1
        if self.word_window_size > 1:
            padded_sentence = np.concatenate((self.pre_padding,
                                              sentence,
                                              self.pos_padding))
        else:
            padded_sentence = sentence
        
        for i in range(self.word_window_size):
            
            for j, table in enumerate(self.feature_tables):
                # this is the column for the i-th position in the window
                # regarding features from the j-th table
                table_deltas = self.input_deltas[:, start_from:start_from + table.shape[1]]
                start_from += table.shape[1]
                
                for token, deltas in zip(padded_sentence[i:], table_deltas):
                    table[token[j]] += deltas
            
            dist_deltas = self.target_dist_deltas[:, dist_target_from : dist_target_from + self.target_dist_table.shape[1] ]
            pre_deltas = dist_deltas.take(np.arange(pre_dist), 0).sum(0)
            pos_deltas = dist_deltas.take(np.arange(-pos_dist, 0), 0).sum(0)
            self.target_dist_table[1:-1, :] += dist_deltas[pre_dist : -pos_dist]
            self.target_dist_table[0] += pre_deltas
            self.target_dist_table[-1] += pos_deltas 
            dist_target_from += self.target_dist_table.shape[1]
            
            dist_deltas = self.pred_dist_deltas[:, dist_pred_from : dist_pred_from + self.pred_dist_table.shape[1] ]
            pre_deltas = dist_deltas.take(np.arange(pre_dist), 0).sum(0)
            pos_deltas = dist_deltas.take(np.arange(-pos_dist, 0), 0).sum(0)
            self.pred_dist_table[1:-1, :] += dist_deltas[pre_dist : -pos_dist]
            self.pred_dist_table[0] += pre_deltas
            self.pred_dist_table[-1] += pos_deltas
            
            pre_dist -= 1
            pos_dist += 1
            dist_pred_from += self.pred_dist_table.shape[1]
            
        self._create_target_lookup()
        self._create_pred_lookup()
    
    @cython.boundscheck(False)
    def _viterbi(self, np.ndarray[FLOAT_t, ndim=2] scores, bool allow_repeats=True):
        """
        Performs a Viterbi search over the scores for each tag using
        the transitions matrix. If a matrix wasn't supplied, 
        it will return the tags with the highest scores individually.
        """
        if self.transitions is None:
            best_scores = scores.argmax(1)
            
            if allow_repeats:
                return best_scores
            
            # we must find the combination of tags that maximizes the probabilities
            logadd = np.log(np.sum(np.exp(scores), 1))
            logprobs = (scores.T - logadd).T
            counts = np.bincount(best_scores)
            
            while counts.max() != 1:
                # find the tag with the most conflicting args
                conflicting_tag = counts.argmax()
                
                # arguments with that tag as current maximum
                args = np.where(best_scores == conflicting_tag)[0]
                
                # get the logprobs for those args having this tag
                conflicting_probs = logprobs[args, conflicting_tag]
                
                # find the argument with the highest probability for that tag
                highest_prob_arg = args[conflicting_probs.argmax()] 
                
                # set the score for other arguments in that tag to a low value
                other_args = args[args != highest_prob_arg]
                scores[other_args, conflicting_tag] = -1000
                
                # and find the new maxes, without recalculating probabilities
                best_scores = scores.argmax(1)
                counts = np.bincount(best_scores)
            
            return best_scores
        
        path_scores = np.empty_like(scores)
        path_backtrack = np.empty_like(scores, int)
        #path_backtrack = np.empty_like(scores, np.int)
        
        # now the actual Viterbi algorithm
        # first, get the scores for each tag at token 0
        # the last row of the transitions table has the scores for the first tag
        path_scores[0] = scores[0] + self.transitions[-1]
        
        for i, token in enumerate(scores[1:], 1):
            
            # each line contains the score until each tag t plus the transition to each other tag t'
            prev_score_and_trans = (path_scores[i - 1] + self.transitions[:-1].T).T
            
            # find the previous tag that yielded the max score
            path_backtrack[i] = prev_score_and_trans.argmax(0)
            path_scores[i] = prev_score_and_trans[path_backtrack[i], 
                                                  np.arange(self.output_size)] + scores[i]
            
        # now find the maximum score for the last token and follow the backtrack
        answer = np.empty(len(scores), dtype=int)
        #answer = np.empty(len(scores), dtype=np.int)
        answer[-1] = path_scores[-1].argmax()
        self.answer_score = path_scores[-1][answer[-1]]
        previous_tag = path_backtrack[-1][answer[-1]]
        
        for i in range(len(scores) - 2, 0, -1):
            answer[i] = previous_tag
            previous_tag = path_backtrack[i][previous_tag]
        
        answer[0] = previous_tag
        return answer
    
    def _create_target_lookup(self):
        """
        Creates a lookup table with the window value for each different distance
        to the target token (target_dist_lookup) and one with the precomputed
        values in the convolution layer (target_convolution_lookup). 
        """
        # consider padding. if the table has 10 entries, with a word window of 3,
        # we would have to consider up to the distance of 11, because of the padding.
        num_distances = self.target_dist_table.shape[0] + self.word_window_size - 1
        self.target_dist_lookup = np.empty((num_distances, 
                                            self.word_window_size * self.target_dist_table.shape[1]))
        self.target_dist_offset = num_distances / 2
        window_from = 0
        window_to = self.target_dist_table.shape[1] 
        for i in range(self.word_window_size):
            # each token in the window will is shifted in relation to the middle one
            shift = i - self.half_window
            
            # discount half window size because of the extra distances we added for padding
            inds = np.arange(shift, num_distances + shift) - self.half_window
            inds = np.clip(inds, 0, self.target_dist_table.shape[0] - 1)
            self.target_dist_lookup[:,window_from : window_to] = self.target_dist_table[inds,]
            
            window_from = window_to
            window_to += self.target_dist_table.shape[1]
        
        self.target_convolution_lookup = self.target_dist_lookup.dot(self.target_dist_weights)
        
    
    def _create_pred_lookup(self):
        """
        Creates a lookup table with the window value for each different distance
        to the predicate token (pred_dist_lookup) and one with the precomputed
        values in the convolution layer (pred_convolution_lookup). 
        """
        # consider padding. if the table has 10 entries, with a word window of 3,
        # we would have to consider up to the distance of 11, because of the padding.
        num_distances = self.pred_dist_table.shape[0] + self.word_window_size - 1
        self.pred_dist_lookup = np.empty((num_distances, 
                                          self.word_window_size * self.pred_dist_table.shape[1]))
        self.pred_dist_offset = num_distances / 2
        window_from = 0
        window_to = self.pred_dist_table.shape[1] 
        for i in range(self.word_window_size):
            # each token in the window will is shifted in relation to the middle one
            shift = i - self.half_window
            
            # discount half window size because of the extra distances we added for padding
            inds = np.arange(shift, num_distances + shift) - self.half_window
            inds = np.clip(inds, 0, self.pred_dist_table.shape[0] - 1)
            self.pred_dist_lookup[:,window_from : window_to] = self.pred_dist_table[inds,]
            
            window_from = window_to
            window_to += self.pred_dist_table.shape[1] 
        
        self.pred_convolution_lookup = self.pred_dist_lookup.dot(self.pred_dist_weights)
    
    def _create_convolution_lookup(self, sentence, training):
        """
        Creates a lookup table storing the values found by each
        convolutional neuron before summing distance features.
        The table has the format len(sent) x len(convol layer)
        Biases are not included.
        """
        cdef np.ndarray padded_sentence
        
        # add padding to the sentence
        if self.word_window_size > 1:
            padded_sentence = np.vstack((self.pre_padding,
                                         sentence,
                                         self.pos_padding))
        else:
            padded_sentence = sentence
        
        self.convolution_lookup = np.empty((len(sentence), self.hidden_size))
        #print(len(sentence), self.hidden_size, len(self.convolution_lookup[0]))
        
        # first window
        cdef np.ndarray window = padded_sentence[:self.word_window_size]
        cdef np.ndarray input_data
        input_data = np.concatenate(
                                    [table[index] 
                                     for token_indices in window
                                     for index, table in zip(token_indices, 
                                                             self.feature_tables)
                                     ]
                                    )
        self.convolution_lookup[0] = self.hidden_weights.dot(input_data)
        if training:
            # store the values of each input -- needed when adjusting features
            self.input_sent_values[0] = input_data
        
        cdef np.ndarray[FLOAT_t, ndim=1] new_data
        for i, element in enumerate(padded_sentence[self.word_window_size:], 1):
            new_data = np.concatenate([table[index] for 
                                       index, table in zip(element, self.feature_tables)])
            
            # slide the window to the next element
            input_data = np.concatenate((input_data[self.features_per_token:], 
                                         new_data))
            
            #print(i, input_data)
            self.convolution_lookup[i] = self.hidden_weights.dot(input_data)
            if training:
                self.input_sent_values[i] = input_data
        



import numpy as np

"""
source code = https://visualstudiomagazine.com/articles/2021/08/16/wasserstein-distance.aspx
By James McCaffrey
"""

from nltkor.make_requirement import make_requirement
try:
    import torch
except ImportError:
    requirement = ['torch']
    file_path = make_requirement(requirement)
    raise Exception(f"""
    Need to install Libraries, please pip install below libraries
    \t pip install torch
    Or, use pip install requirement.txt
    \t  pip install -r {file_path}
    """)


class WassersteinDistance:
    def __init__(self) -> None:
        pass

    def first_nonzero(self, vec):
        dim = len(vec)
        for i in range(dim):
            if vec[i] > 0.0:
                return i
        return -1  # no empty cells found

    def move_dirt(self, dirt, di, holes, hi):
    # move as much dirt at [di] as possible to h[hi]
        if dirt[di] <= holes[hi]:   # use all dirt
            flow = dirt[di]
            dirt[di] = 0.0            # all dirt got moved
            holes[hi] -= flow         # less to fill now
        elif dirt[di] > holes[hi]:  # use just part of dirt
            flow = holes[hi]          # fill remainder of hole
            dirt[di] -= flow          # less dirt left
            holes[hi] = 0.0           # hole is filled
        dist = np.abs(di - hi)
        return flow * dist          # work

    def compute_wasserstein(self, p, q):
        if "torch" in str(type(p)):
            p = p.numpy()
        if "torch" in str(type(q)):
            q = q.numpy()

        dirt = np.copy(p)
        holes = np.copy(q)
        tot_work = 0.0

        while True:  # TODO: add sanity counter check
            from_idx = self.first_nonzero(dirt)
            to_idx = self.first_nonzero(holes)
            if from_idx == -1 or to_idx == -1:
                break
            work = self.move_dirt(dirt, from_idx, holes, to_idx)
            tot_work += work
        return tot_work

    def kullback_leibler(self, p, q):
        n = len(p)
        sum = 0.0
        for i in range(n):
            sum += p[i] * np.log(p[i] / q[i])
        return sum

    def compute_kullback(self, p, q):
        if "torch" in str(type(p)):
            p = p.numpy()
        if "torch" in str(type(q)):
            q = q.numpy()
        a = self.kullback_leibler(p, q)
        b = self.kullback_leibler(q, p)
        return a + b

    def compute_jesson_shannon(self, p, q):
        if "torch" in str(type(p)):
            p = p.numpy()
        if "torch" in str(type(q)):
            q = q.numpy()

        a = self.kullback_leibler(p, (p + q)/2)
        b = self.kullback_leibler(q, (p + q)/2)
        return (a + b)/2


def demo():
    print("\nBegin Wasserstein distance demo ")

    P =  np.array([0.6, 0.1, 0.1, 0.1, 0.1])
    Q1 = np.array([0.1, 0.1, 0.6, 0.1, 0.1])
    Q2 = np.array([0.1, 0.1, 0.1, 0.1, 0.6])

    P = torch.from_numpy(P)
    Q1 = torch.from_numpy(Q1)
    Q2 = torch.from_numpy(Q2)
    kl_p_q1 = WassersteinDistance().compute_kullback(P, Q1)
    kl_p_q2 = WassersteinDistance().compute_kullback(P, Q2)

    wass_p_q1 = WassersteinDistance().compute_wasserstein(P, Q1)
    wass_p_q2 = WassersteinDistance().compute_wasserstein(P, Q2)

    jesson_p_q1 = WassersteinDistance().compute_jesson_shannon(P, Q1)
    jesson_p_q2 = WassersteinDistance().compute_jesson_shannon(P, Q2)

    print("\nKullback-Leibler distances: ")
    print("P to Q1 : %0.4f " % kl_p_q1)
    print("P to Q2 : %0.4f " % kl_p_q2)

    print("\nWasserstein distances: ")
    print("P to Q1 : %0.4f " % wass_p_q1)
    print("P to Q2 : %0.4f " % wass_p_q2)

    print("\nJesson-Shannon distances: ")
    print("P to Q1 : %0.4f " % jesson_p_q1)
    print("P to Q2 : %0.4f " % jesson_p_q2)

    print("\nEnd demo ")

if __name__ == "__main__":
    demo()



import os
import unicodedata
import sys


for filename in os.listdir('/01. 체언_상세//'):
	new_filename=filename.replcace(filename, unicodedata.normalize('NFD',filename))
	os.rename(filename, new_filename)







# -*- coding: utf-8 -*-

import logging
import numpy as np

from .word_dictionary import WordDictionary as WD
from collections import defaultdict

# dummy value to be used when POS is an additional attribute
PADDING_POS = 'PADDING'

class Caps(object):
    """Dummy class for storing numeric values for capitalization."""
    num_values = 5
    lower = 0
    title = 1
    non_alpha = 2
    other = 3
    padding = 4


class Token(object):
    def __init__(self, word, morph_h='NA', pos_h='NA',  morph_t='NA', pos_t='NA', chunk='NA'):
        """
        A token representation that stores discrete attributes to be given as 
        input to the neural network. 
        """
        self.word = word
        self.morph_h = morph_h
        self.pos_h = pos_h
        self.pos_t = pos_t
        self.morph_t = morph_t
        self.chunk = chunk
    
    def __str__(self):
        return str(self.word)
    
    def __repr__(self):
        return self.word.__repr__()


class Affix(object):
    """Dummy class for manipulating suffixes and their related codes."""
    # codes maps integers (affix sizes) to dicts. each dict maps a suffix of the given 
    # size to its code
    suffix_codes = {}
    prefix_codes = {}
    other = 0
    padding = 1
    num_suffixes_per_size = {}
    num_prefixes_per_size = {}
    
    @classmethod
    def load_suffixes(cls, md):
        """
        Loads suffixes from the suffix file.
        """
        cls.load_affixes(cls.suffix_codes, md.paths['suffixes'])
        
        # +2 because of the unkown suffix code and padding
        cls.num_suffixes_per_size = {size: len(cls.suffix_codes[size]) + 2
                                     for size in cls.suffix_codes}
    
    @classmethod
    def load_prefixes(cls, md):
        """
        Loads prefixes from the prefix file.
        """
        cls.load_affixes(cls.prefix_codes, md.paths['prefixes'])
        
        # +2 because of the unkown prefix code and padding
        cls.num_prefixes_per_size = {size: len(cls.prefix_codes[size]) + 2
                                     for size in cls.prefix_codes}
        
    
    @classmethod
    def load_affixes(cls, codes, filename):
        """
        Parent function for loading prefixes and suffixes.
        """
        logger = logging.getLogger("Logger")
        
        # intermediate storage
        affixes_by_size = defaultdict(list)
        
        try:
            with open(filename, 'rb') as f:
                for line in f:
                    affix = unicode(line.strip(), 'utf-8')
                    size = len(affix)
                    affixes_by_size[size].append(affix)
        except IOError:
            logger.error("File %s doesn't exist." % filename)
            raise
        
        for size in affixes_by_size:
            # for each size, each affix has a code starting from 2
            # 0 is reserved for unknown affixes
            # 1 is reserved for padding pseudo-affixes
            codes[size] = {affix: code 
                           for code, affix in enumerate(affixes_by_size[size], 2)}
    
    @classmethod
    def get_suffix(cls, word, size):
        """
        Return the suffix code for the given word. Consider a suffix
        of the given size.
        """
        if word == WD.padding_left or word == WD.padding_right:
            return cls.padding
        
        if len(word) <= size:
            return cls.other
        
        suffix = word[-size:].lower()
        code = cls.suffix_codes[size].get(suffix, cls.other)
        return code
    
    @classmethod
    def get_prefix(cls, word, size):
        """
        Return the suffix code for the given word. Consider a suffix
        of the given size.
        """
        if word == WD.padding_left or word == WD.padding_right:
            return cls.padding
        
        if len(word) <= size:
            return cls.other
        
        prefix = word[:size].lower()
        code = cls.prefix_codes[size].get(prefix, cls.other)
        return code


class TokenConverter(object):
    
    def __init__(self):
        """
        Class to convert tokens into indices to their feature vectos in
        feature matrices.
        """
        self.extractors = []
    
    def add_extractor(self, extractor):
        """
        Adds an extractor function to the TokenConverter. In order to get a token's 
        feature indices, the Converter will call each of its extraction functions passing
        the token as a parameter. The result will be a list containing each result. 
        """
        self.extractors.append(extractor)
    
    def get_padding_left(self, tokens_as_string=True):
        """
        Returns an object to be used as the left padding in the sentence.
        
        :param tokens_as_string: if True, treat tokens as strings. 
        If False, treat them as Token objects.
        """
        if tokens_as_string:
            pad = WD.padding_left
        else:
            pad = Token(WD.padding_left, morph_h=WD.padding_left, \
                morph_t=WD.padding_left, pos_h=PADDING_POS,  pos_t=PADDING_POS)
        return self.convert(pad)
    
    def get_padding_right(self, tokens_as_string=True):
        """
        Returns an object to be used as the right padding in the sentence.
        
        :param tokens_as_string: if True, treat tokens as strings. 
            If False, treat them as Token objects.
        """
        if tokens_as_string:
            pad = WD.padding_right
        else:
            pad = Token(WD.padding_right, morph_h=WD.padding_right, \
                morph_t=WD.padding_right, pos_h=PADDING_POS, pos_t=PADDING_POS)
            #pad = Token(WD.padding_right, pos_t=PADDING_POS)
        return self.convert(pad)
    
    def convert(self, token):
        """
        Converts a token into its feature indices.
        """
        indices = np.array([function(token) for function in self.extractors])
        return indices


def get_capitalization(word):
    """
    Returns a code describing the capitalization of the word:
    lower, title, other or non-alpha (numbers and other tokens that can't be
    capitalized).
    """
    if word == WD.padding_left or word == WD.padding_right:
        return Caps.padding
    
    if not any(c.isalpha() for c in word):
        # check if there is at least one letter
        # (this is faster than using a regex)
        return Caps.non_alpha
    
    if word.islower():
        return Caps.lower
    
    # word.istitle() returns false for compunds like Low-cost
    if len(word) == 1:
        # if we reached here, there's a single upper case letter
        return Caps.title
    elif word[0].isupper() and word[1:].islower():
        return Caps.title    
    
    return Caps.other

def capitalize(word, capitalization):
    """
    Capitalizes the word in the desired format. If the capitalization is 
    Caps.other, it is set all uppercase.
    """
    if capitalization == Caps.non_alpha or capitalization == Caps.padding:
        return word
    elif capitalization == Caps.lower:
        return word.lower()
    elif capitalization == Caps.title:
        return word[0].upper() + word[1:].lower()
    elif capitalization == Caps.other:
        return word.upper()
    else:
        raise ValueError("Unknown capitalization type.")




# -*- coding: utf-8 -*-

"""
A convolutional neural network for NLP tagging tasks such as dependency
parsing, where each token has another (or root) as a head. 
"""

import numpy as np
cimport numpy as np

cdef class ConvolutionalDependencyNetwork(ConvolutionalNetwork):
    
    # the weights of all possible dependency among tokens
    cdef readonly np.ndarray dependency_weights
    
    # validation data
    cdef validation_heads
    
    def save(self):
        """
        Saves the neural network to a file.
        It will save the weights, biases, sizes, padding and 
        distance tables, and other feature tables.
        """
        data = self._generate_save_dict()
        
        np.savez(self.network_filename, **data)
    
    @classmethod
    def load_from_file(cls, filename):
        """
        Loads the neural network from a file.
        It will load weights, biases, sizes, padding and 
        distance tables, and other feature tables.
        """
        data = np.load(filename, allow_pickle=True)
        nn = cls._load_from_file(data, filename)
        
        return nn
    
    def train(self, list sentences, list heads, int epochs, 
              int epochs_between_reports=0, float desired_accuracy=0,
              list labels=None):
        """
        Trains the convolutional network. Refer to the basic Network
        train method for detailed explanation. 
        """
        # the ConvolutionalNetwork class was written primarily for SRL
        # every token acts as a predicate, and we don't need to tell it explicitely
        predicates = [np.arange(len(sentence)) for sentence in sentences]
        
        # the last argument in ConvolutionalNetwork.train is actually the argument
        # groups list. We use "labels" here just to signal that there is non-None
        # argument, which is correctly handled by the DependencyNetwork._tag(...) method.
        
        if self.validation_sentences is None:
            self.set_validation_data(sentences, heads, labels)
        
        super(ConvolutionalDependencyNetwork, self).train(sentences, predicates, 
                                                          heads, epochs, 
                                                          epochs_between_reports, 
                                                          desired_accuracy,
                                                          labels)
        
    def set_validation_data(self, list sentences, list heads, list labels=None):
        """
        Sets the data to be used in validation during training. If this function
        is not called before training, the training data itself is used to 
        measure the model's performance.
        
        :param labels: only used when learning labels
        """
        self.validation_sentences = sentences
        self.validation_tags = labels
        self.validation_heads = heads
    
    def _tag_sentence(self, sentence, predicates=None, heads=None, labels=None):
        """
        This function is just an interface to the _tag_sentence signature
        defined in ConvolutionalNetwork.
        """
        if labels is None:
            self._tag_sentence_unlabeled_dependency(sentence, heads)
        else:
            self._tag_sentence_labeled_dependency(sentence, heads, labels)
    
    def _tag_sentence_unlabeled_dependency(self, np.ndarray sentence, np.ndarray heads=None):
        """
        Run the network for the unlabeled dependency task.
        A graph with all weights for possible dependencies is built 
        and the final answer is obtained applying the Chu-Liu-Edmond's
        algorithm.
        """
        training = heads is not None
        self._pre_tagging_setup(sentence, training)
        
        num_tokens = len(sentence)
        # dependency_weights [i, j] has the score for token i having j as a head.
        # the main diagonal has the values for dependencies from the root and is 
        # later copied to the last column for easier processing
        self.dependency_weights = np.empty((num_tokens, num_tokens + 1))
        
        cdef np.ndarray[FLOAT_t, ndim=1] token_scores
        
        # in the SRL parlance, each token is treated as a predicate, because all
        # sentence tokens are scored with respect to it (in order to determine the
        # dependency weights)
        for token in range(num_tokens):
            
            # _sentence_convolution returns a 2-dim array. in dep parsing, 
            # we only have one dimension, so reshape it
            token_scores = self._sentence_convolution(sentence, token, 
                                                      training=training).reshape(num_tokens)
            self.dependency_weights[token, :-1] = token_scores
            
            if training:
                head = heads[token]
                 
                if self._calculate_gradients(head, token_scores):
                    self._backpropagate()
                    self._calculate_input_deltas(sentence, token)
                    self._adjust_weights(token)
                    self._adjust_features(sentence, token)
        
        # copy dependency weights from the root to each token to the last column and
        # effectively ignore the main diagonal (dependency to the token itself)
        self.dependency_weights[np.arange(num_tokens), 
                                -1] = self.dependency_weights.diagonal()
        np.fill_diagonal(self.dependency_weights, -np.Infinity)
        answer = self._find_maximum_spanning_tree()
        
        return answer
    
    def _tag_sentence_labeled_dependency(self, np.ndarray sentence, np.ndarray heads,
                                         np.ndarray labels=None):
        """
        Run the network for labeling pre determined dependency edges between tokens.
        This is similar to the classification step in SRL.
        """
        cdef np.ndarray[FLOAT_t, ndim=1] answer
        cdef np.ndarray[FLOAT_t, ndim=2] scores
        training = labels is not None
        self._pre_tagging_setup(sentence, training)
        
        answer = np.zeros(len(sentence))
        
        # as in unlabeled dependency, each token is treated as a predicate from the
        # SRL point of view. The only target is its head
        for token in range(len(sentence)):
            head = heads[token]
            
            # weird format just to take advantage of the SRL classification code
            # it means that the target starts at position *head* and ends at *head*
            head = [[head, head]]
            
            # it will return a 2-dim array, but we only have one target
            # argmax() works as expected
            scores = self._sentence_convolution(sentence, token, head, training)
            answer[token] = scores.argmax()
            
            if training:
                label = labels[token]
                if self._calculate_gradients_classify([label], scores):
                    self._backpropagate()
                    self._calculate_input_deltas(sentence, token, head)
                    self._adjust_weights(token, head)
                    self._adjust_features(sentence, token)
        
        
        return answer
    
    def _calculate_gradients(self, gold_head, scores):
        """
        Calculate the gradients to be applied in the backpropagation. Gradients
        are calculated after the network has output the scores for assigning 
        each token as head of a given token. 
        
        We aim at maximizing the log probability of the right head:
        log(p(head)) = score(head) - logadd(scores for all heads)
        
        :param gold_head: the index of the token that should have the highest 
            score
        :param scores: the scores output by the network
        :returns: if True, normal gradient calculation was performed.
            If False, the error was too low and weight correction should be
            skipped. 
        """
        # first, set the gradient at each token to 
        # -exp(score(token)) / sum_j exp(score(token_j))
        # i.e., the negative of its probability
        cdef np.ndarray[FLOAT_t, ndim=1] exp_scores = np.exp(scores)
        exp_sum = np.sum(exp_scores)
        self.net_gradients = -exp_scores / exp_sum
        error = 1 + self.net_gradients[gold_head]
        self.error += error
        
        # check if the error is too small - if so, not worth to continue
        if error <= 0.01:
            self.skips += 1
            return False
        
        # and add 1 to the right head
        self.net_gradients[gold_head] += 1
        
        # the ConvolutionalNetwork class deals with multi dimensional gradients
        # (because of more than one output neuron), so let's reshape
        new_shape = (self.net_gradients.shape[0], 1)
        self.net_gradients = self.net_gradients.reshape(new_shape)  
        
        return True
    
    def _validate(self):
        """
        Evaluate the network performance by token hit and whole sentence hit.
        """
        hits = 0
        num_tokens = 0
        sentence_hits = 0
        
        for i in range(len(self.validation_sentences)):
            sent = self.validation_sentences[i]
            heads = self.validation_heads[i]
            sentence_hit = True
            
            if self.validation_tags is None:
                # unlabeled dependency
                answer = self._tag_sentence_unlabeled_dependency(sent)
                gold_tags = heads
            else:
                # labeled dependency
                gold_tags = self.validation_tags[i]
                answer = self._tag_sentence_labeled_dependency(sent, heads)
                
            for j in range(len(gold_tags)):
                net_tag = answer[j]
                gold_tag = gold_tags[j]
                
                if net_tag == gold_tag or (gold_tag == j and net_tag == len(sent)):
                    hits += 1
                else:
                    sentence_hit = False
            
            if sentence_hit:
                sentence_hits += 1
            num_tokens += len(sent)
        
        self.accuracy = float(hits) / num_tokens
        self.sentence_accuracy = float(sentence_hits) / len(self.validation_sentences)
        
    
    def _average_error(self):
        """
        Average the network error over tokens.
        """
        self.error = self.error / self.num_tokens
    
    def _print_epoch_report(self, int num):
        """
        Reports the status of the network in the given training
        epoch, including error, token and sentence accuracy.
        """
        logger = logging.getLogger("Logger")
        logger.info("%d epochs   Error: %f   Token accuracy: %f   " \
                    "Sentence accuracy: %f    " \
                    "%d corrections skipped   " \
                    "Learning rate: %f"  % (num,
                                            self.error,
                                            self.accuracy,
                                            self.sentence_accuracy,
                                            self.skips,
                                            self.learning_rate))
    
    def tag_sentence(self, np.ndarray sentence, np.ndarray heads=None):
        """
        If heads is not given, compute the dependency edges in the sentence.
        If it is given, compute the label of each dependency.
        
        :returns: a numpy 1-dim array with the head for each token or label
            of each edge. In the first case, a dependency from the root is 
            represented as a value equal to the sentence length. 
        """
        if heads is None:
            return self._tag_sentence_unlabeled_dependency(sentence)
        else:
            return self._tag_sentence_labeled_dependency(sentence, heads)
    
    def _find_cycles(self, np.ndarray graph):
        """
        Check if the given graph has cycles and returns the first one
        to be found.
        
        :param graph: an array where graph[i] has the number of a
            vertex with an incoming connection to i
        """
        # this set stores all vertices with a valid path to the root
        reachable_vertices = set()
        
        # vertices known to be unreachable from the root, i.e., in a cycle
        vertices_in_cycles = set()
        
        # vertices currently being evaluated, not known if they're reachable
        visited = set()
        
        cycles = []
        
        # the directions of the edges don't matter if we only want to find cycles
        for vertex in range(len(graph)):
            if vertex in reachable_vertices or vertex in vertices_in_cycles:
                continue
            
            cycle = self._find_cycle_recursive(graph, vertex, visited, 
                                               reachable_vertices, vertices_in_cycles)
            if cycle is not None:
                cycles.append(cycle)
        
        return cycles
            
    def _find_cycle_recursive(self, np.ndarray graph, int vertex, set visited, 
                              set reachable, set unreachable):
        """
        Auxiliary recursive function for searching the graph for cycles.
        It returns the first cycle it can find starting from the given
        vertex, or None.
        """
        next_vertex = graph[vertex]
        root = len(graph)
        visited.add(vertex)
        
        if next_vertex == root or next_vertex in reachable:
            # vertex linked to root
            reachable.update(visited)
            visited.clear()
            cycle = None
        
        elif next_vertex in visited:
            # cycle detected! return all vertices in it
            visited.clear()
            cycle = set([vertex])
            while next_vertex != vertex:
                cycle.add(next_vertex)
                next_vertex = graph[next_vertex]
            
            unreachable.update(cycle)
        
        elif next_vertex in unreachable:
            # vertex linked to an existing cycle, but not part of it
            # (if it were, it should have been filtered out in _find_cycles)
            visited.clear()
            cycle = None
        
        else:
            # continue checking
            cycle = self._find_cycle_recursive(graph, next_vertex, visited, 
                                               reachable, unreachable)
        
        return cycle
    
    def _contract_cycle(self, heads, cycle):
        """
        Contract the given cycle in the dependency graph.
        
        :param heads: list of the heads of each token, such that
            heads[i] contains the head for the i-th token
        :param cycle: a set containing the numbers of the vertices
            in the cycle
        """
        # each cell i, j has the score for token i having j as its head.
        
        num_vertices = self.dependency_weights.shape[1]
        outside = np.array([x for x in range(num_vertices) if x not in cycle])
        cycle = np.array(list(cycle))
        
        cdef np.ndarray[FLOAT_t, ndim=2] outgoing_weights
        
        # adjustments will be made on incoming and outgoing edges
        # pick the part of the weight matrix that contain them
        
        # if len(outside) == 1, it means all vertices except for the root are in a cycle
        if len(outside) > 1:
            # weird index array we need in order to properly use fancy indexing
            # -1 because we can't take the root now
            outside_inds = np.array([[i] for i in outside[:-1]])
            outgoing_weights = self.dependency_weights[outside_inds, cycle]
            
            # the cycle should have only one outgoing edge for each vertex outside it
            # so, search the maximum outgoing edge to each outside vertex
            max_outgoing_inds = outgoing_weights.argmax(1)
            max_outgoing_weights = outgoing_weights.max(1)
            
            # set every outgoing weight to -inf and then restore the highest ones
            outgoing_weights[:] = -np.Infinity
            outgoing_weights[np.arange(len(outside_inds)), 
                             max_outgoing_inds] = max_outgoing_weights
            self.dependency_weights[outside_inds, cycle] = outgoing_weights
        
        # now, adjust incoming edges. the incoming edge from each vertex v 
        # (outside the cycle) to v' (inside) is reweighted as: 
        # s(v, v') = s(v, v') - s(head(v'), v') + s(c)
        # and then we pick the highest edge for each outside vertex
        # s(c) = sum_v s(head(v), v)
        cycle_inds = np.array([[i] for i in cycle])
        cdef np.ndarray[FLOAT_t, ndim=2] incoming_weights
        incoming_weights = self.dependency_weights[cycle_inds, outside]
        
        cycle_score = 0
        for i, vertex in enumerate(cycle):
            head_to_v = self.dependency_weights[vertex, heads[vertex]]
            cycle_score += head_to_v
            incoming_weights[i] -= head_to_v
        
        max_incoming_inds = incoming_weights.argmax(0)
        max_incoming_weights = incoming_weights.max(0)
        # we leave the + s(c) to the end
        max_incoming_weights += cycle_score
        
        # the vertex with the maximum weighted incoming edge now changes
        # its head, thus breaking the cycle
        new_head_ind = max_incoming_weights.argmax()
        vertex_leaving_cycle_ind = max_incoming_inds[new_head_ind]
        
        new_head = outside[new_head_ind]
        vertex_leaving_cycle = cycle[vertex_leaving_cycle_ind]
        old_head = heads[vertex_leaving_cycle]
        heads[vertex_leaving_cycle] = new_head
        self.dependency_weights[vertex_leaving_cycle, old_head] = -np.Infinity
        
        # analagous to the outgoing weights
        incoming_weights[:] = -np.Infinity
        incoming_weights[max_incoming_inds,
                         np.arange(len(outside))] = max_incoming_weights
        self.dependency_weights[cycle_inds, outside] = incoming_weights
    
    def _find_maximum_spanning_tree(self):
        """
        Run the Chu-Liu / Edmond's algorithm in order to find the highest
        scoring dependency tree from the dependency graph weights.
        
        :returns: a 1-dim array with the head of each token in the sentence
        """
        # pick the highest scoring dependency for each word
        heads = self.dependency_weights.argmax(1)
        
        # check if there are cycles. if there isn't any, we're done
        cycles = self._find_cycles(heads)
        
        for cycle in cycles:
            # resolve each cycle c
            self._contract_cycle(heads, cycle)
        
        return heads
    



# -*- coding: utf-8 -*-

"""
Auxiliary functions for SRL training.
"""

import re
import numpy as np


def init_transitions_simplified(tag_dict):
    """
    This function initializes a tag transition table containing only
    the boundaries IOBES.
    """
    tags = sorted(tag_dict, key=tag_dict.get)
    transitions = []
    
    for tag in tags:
        if tag in 'OES':
            trans = lambda x: 0 if x in 'BOS' else -1000
        elif tag in 'IB':
            trans = lambda x: 0 if x in 'IE' else -1000
        else:
            raise ValueError('Unexpected tag: %s' % tag)
        
        transitions.append([trans(next_tag) for next_tag in tags])
    
    # initial transition
    trans = lambda x: 0 if x in 'BOS' else -1000
    transitions.append([trans(next_tag) for next_tag in tags])
    
    return np.array(transitions, np.float)


def init_transitions(tag_dict, scheme):
    """
    This function initializes the tag transition table setting 
    very low values for impossible transitions.
     
    :param tag_dict: The tag dictionary mapping tag names to the
        network output number.
    :param scheme: either iob or iobes.
    """
    scheme = scheme.lower()
    assert scheme in ('iob', 'iobes'), 'Unknown tagging scheme: %s' % scheme
    transitions = []
    
    # since dict's are unordered, let's take the tags in the correct order
    tags = sorted(tag_dict, key=tag_dict.get)
    
    # transitions between tags
    for tag in tags:
        
        if tag == 'O':
            # next tag can be O, V or any B
            trans = lambda x: 0 if re.match('B|S|V', x) \
                                else -1 if x == 'O' else -1000
        
        elif tag[0] in 'IB':
            block = tag[2:]
            if scheme == 'iobes':
                # next tag can be I or E (same block)
                trans = lambda x: 0 if re.match('(I|E)-%s' % block, x) else -1000
            else:
                # next tag can be O, I (same block) or B (new block)
                trans = lambda x: 0 if re.match('I-%s' % block, x) or re.match('B-(?!%s)' % block, x) \
                                    else -1 if x == 'O' else -1000
        
        elif tag[0] in 'ES':
            # next tag can be O, S (new block) or B (new block)
            block = tag[2:]
            trans = lambda x: 0 if re.match('(S|B)-(?!%s)' % block, x) \
                                else -1 if x == 'O' else -1000

        else:
            raise ValueError('Unknown tag: %s' % tag)
        
        transitions.append([trans(next_tag) for next_tag in tags])  
    
    # starting tag
    # it can be O or any B/S
    trans = lambda x: 0 if x[0] in 'OBS' else -1000
    transitions.append([trans(next_tag) for next_tag in tags])
    
    return np.array(transitions, np.float)




"""
string2string alignment
src = https://github.com/stanfordnlp/string2string


MIT License

Copyright (c) 2023 Mirac Suzgun

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


"""

import multiprocessing

from typing import List, Union, Tuple, Optional
import numpy as np
from nltkor.make_requirement import make_requirement

try:
    from joblib import Parallel, delayed
    from tqdm import tqdm
except ImportError:
    requirement = ['joblib', 'tqdm>=4.40.0']
    file_path = make_requirement(requirement)
    raise Exception(f"""
    Need to install Libraries, please pip install below libraries
    \t pip install joblib
    \t pip install tqdm>=4.40.0
    Or, use pip install requirement.txt
    \t  pip install -r {file_path}
    """)


# for dev purposes
import sys
# sys.path.append("/Users/dowon/nltk_ko/nltk/misc")
from nltkor.misc.string2string_basic_functions import cartesian_product
# from string2string_basic_functions import cartesian_product


# Parent class for all alignment algorithms
class StringAlignment:
    """
        This class is the parent class for all alignment algorithms implemented in this module.
    """
    # Initialize the class.
    def __init__(self,
        match_weight: int = 1.,
        mismatch_weight: int = -1.,
        gap_weight: int = -1,
        gap_char: str = "-",
        match_dict: dict = None,
        ) -> None:
        r"""
        This function initializes the StringAlignment class.

        Arguments:
            match_weight (int): The weight for a match (default: 1).
            mismatch_weight (int): The weight for a mismatch (default: -1).
            gap_weight (int): The weight for a gap (default: -1).
            gap_char (str): The character for a gap (default: "-").
            match_dict (dict): The match dictionary (default: None).

        Returns:
            None

        .. note::

            The match_dict represents a dictionary of the match weights for each pair of characters. For example, if the match_dict is {"A": {"A": 1, "T": -1}, "T": {"A": -1, "T": 1}}, then the match weight for "A" and "A" is 1, the match weight for "A" and "T" is -1, the match weight for "T" and "A" is -1, and the match weight for "T" and "T" is 1.
            The match_dict is particularly useful when we wish to align (or match) non-identical characters. For example, if we wish to align "A" and "T", we can set the match_dict to {"A": {"T": 1}}. This will ensure that the match weight for "A" and "T" is 1, and the match weight for "A" and "A" and "T" and "T" is 0.
        """
        # Set the weights.
        self.match_weight = match_weight
        self.mismatch_weight = mismatch_weight
        self.gap_weight = gap_weight
        self.gap_char = gap_char
        self.match_dict = match_dict



    def bool_match(self,
        c1: Union[str, List[str]],
        c2: Union[str, List[str]],
    ) -> bool:
        """
        The function returns whether two characters match, according to the match dictionary (if it exists).

        Arguments:
            c1 (str or list of str): The first character or string.
            c2 (str or list of str): The second character or string.

        Returns:
            Whether the two characters match (True or False)
        """

        # If there is no match dictionary, return whether the characters are the same.
        if self.match_dict is None:
            return c1 == c2
        # Otherwise, return whether the characters match according to the match dictionary.
        else:
            if c1 in self.match_dict and c2 in self.match_dict[c1]:
                return self.match_dict[c1][c2] >= 0
            else:
                return c1 == c2



    # Get the match weight.
    def get_match_weight(self,
        c1: Union[str, List[str]],
        c2: Union[str, List[str]],
    ) -> float:
        """
        This function returns the match weight of two characters.

        Arguments:
            c1 (str or list of str): The first character or string.
            c2 (str or list of str): The second character or string.

        Returns:
            The match weight of the two characters or strings.
        """

        # If there is no match dictionary, return the match weight if the characters are the same, and the mismatch weight otherwise.
        if self.match_dict is None:
            if c1 == c2:
                return self.match_weight
            return self.mismatch_weight
        # Otherwise, return the match weight according to the match dictionary.
        else:
            if c1 in self.match_dict and c2 in self.match_dict[c1]:
                return self.match_dict[c1][c2]
            else:
                if c1 == c2:
                    return self.match_weight
                return self.mismatch_weight



    # Get the gap weight.
    def get_gap_weight(self,
        c: Union[str, List[str]],
    ) -> float:
        """
        This function returns the gap weight of a character or string.

        Arguments:
            c (str or list of str): The character or string.

        Returns:
            The gap weight of the character or string.
        """

        # If there is no match dictionary, return the gap weight.
        if self.match_dict is None:
            return self.gap_weight
        # Otherwise, return the gap weight according to the match dictionary.
        else:
            if c in self.match_dict and self.gap_char in self.match_dict[c]:
                return self.match_dict[c][self.gap_char]
            else:
                return self.gap_weight



    # Get the score of a character pair.
    def get_score(self,
        c1: Union[str, List[str]],
        c2: Union[str, List[str]],
    ) -> float:
        """
        This function returns the score of a character or string pair.

        Arguments:
            c1 (str or list of str): The first character or string.
            c2 (str or list of str): The second character or string.

        Returns:
            The score of the character or string pair.
        """
        # If the characters are the same, return the match weight.
        if c1 == c2:
            return self.match_weight
        # If one of the characters is a gap, return the gap weight.
        elif c1 == self.gap_char or c2 == self.gap_char:
            return self.gap_weight
        # Otherwise, return the mismatch weight.
        else:
            return self.mismatch_weight



    # Get the alignment score of two strings (or list of strings).
    # (This is the sum of the scores of the characters.)
    def get_alignment_score(self,
        str1: Union[str, List[str]],
        str2: Union[str, List[str]],
    ) -> float:
        """
        This function returns the alignment score of two strings (or list of strings).

        Arguments:
            str1 (str or list of str): The first string (or list of strings).
            str2 (str or list of str): The second string (or list of strings).

        Returns:
            The alignment score of the two strings (or list of strings).
        """
        # Get the alignment score by summing the scores of the characters.
        score = 0.
        for c1, c2 in zip(str1, str2):
            score += self.get_score(c1, c2)
        return score



    # Add gaps to the shorter string.
    def add_space_to_shorter(self,
        str1: str,
        str2: str,
    ) -> Tuple[str, str]:
        """
        This function adds gaps to the shorter string to make the two strings the same length.

        Arguments:
            str1 (str): The first string.
            str2 (str): The second string.

        Returns:
            The two strings with the same length.
        """
        # Get the maximum length of the two strings.
        max_len = max(len(str1), len(str2))

        # Pad the shorter string with gaps.
        if len(str1) < max_len:
            str1 = str1 + ' ' * (max_len - len(str1))
        elif len(str2) < max_len:
            str2 = str2 + ' ' * (max_len - len(str2))

        # Return the padded strings.
        return str1, str2



    # Print the alignment.
    def print_alignment(self,
        str1: str,
        str2: str,
    ) -> None:
        """
        This function prints the alignment of two strings.

        Arguments:
            str1 (str): The first string.
            str2 (str): The second string.

        Returns:
            None.
        """
        print(str1)
        print(str2)




    def get_alignment_strings_and_indices(self,
        str1: str,
        str2: str,
        separator: str = ' | ',
    ) -> Tuple[Tuple[List[int], List[int]], List[str], List[str]]:
        """
        This function returns the indices of the aligned characters, and the two strings separated by the separator.

        Arguments:
            str1 (str): The first string, separated by the separator.
            str2 (str): The second string, separated by the separator.
            separator (str): The separator.

        Returns:
            The indices of the aligned characters, and the two strings separated by the separator.

        """
        sym1 = str1.split(separator)
        sym2 = str2.split(separator)

        alignment_indices = []
        # Get the indices of the aligned characters.
        for i in range(len(sym1)):
            if self.bool_match(sym1[i], sym2[i]): #sym1[i] == sym2[i]:
                alignment_indices.append((i, i))

        return alignment_indices, sym1, sym2



    # Compute miltiple pairs in parallel using multiprocessing.
    def compute_multiple_pairs(self,
        pairs: List[Tuple[Union[str, List[str]], Union[str, List[str]]]],
        num_workers: int = 1,
        method: str = "multiprocessing",
        **kwargs,
    ) -> List[Tuple[float, Union[List[str], List[List[str]]]]]:
        """
        This "meta" function computes the alignment score of multiple pairs of strings (or lists of strings) in parallel.

        Arguments:
            pairs (list of tuples): A list of tuples, where each tuple contains two strings (or lists of strings).
            num_workers (int): The number of workers to use for multiprocessing.
            method (str): The method to use for parallelization. Options are "multiprocessing" and "joblib".
            **kwargs: Additional keyword arguments to pass to the compute function.

        Returns:
            A list of tuples, where each tuple contains the alignment score and the alignment of the two strings (or lists of strings), based on the compute function.

        .. note::
            * This function uses multiprocessing, either via the multiprocessing module or via joblib.
            * The multiprocessing module is used by default, but joblib can be used instead by setting method="joblib".
            * We found that joblib is empirically faster than multiprocessing for this particular problem.
        """
        # Compute the alignment score of multiple pairs of strings in parallel.
        if method == "multiprocessing":
            with multiprocessing.Pool(num_workers) as pool:
                results = pool.starmap(
                    self.compute,
                    [(pair[0], pair[1], kwargs) for pair in pairs],
                )
        elif method == "joblib":
            results = Parallel(n_jobs=num_workers)(
                delayed(self.compute)(pair[0], pair[1], **kwargs) for pair in tqdm(pairs)
            )
        else:
            raise ValueError(f"Invalid method: {method}")
        return results



# Needleman-Wunsch algorithm class
class NeedlemanWunsch(StringAlignment):
    # Initialize the class.
    def __init__(self,
        match_weight: float = 1.,
        mismatch_weight: float = -1.,
        gap_weight: float = -1.,
        gap_char: str = "-",
        match_dict: dict = None,
        ) -> None:
        r"""
        This function initializes the Needleman-Wunsch algorithm, which is used to get the global alignment of sequences (e.g., strings or lists of strings) such as DNA sequences.

        The algorithm is described in the following paper: [Needleman1970]_

        Arguments:
            match_weight (float): The weight of a match (default: 1.).
            mismatch_weight (float): The weight of a mismatch (default: -1.).
            gap_weight (float): The weight of a gap (default: -1.).
            match_dict (dict): The match dictionary (default: None).
            gap_char (str): The gap character (default: "-").

        .. [Needleman1970] Needleman, S.B. and Wunsch, C.D., 1970. A General Method Applicable to the Search for Similarities in the Amino Acid Sequence of Two Proteins. Journal of Molecular Biology, 48(3), pp.443-453.
        """
        # Initialize using the parent class.
        super().__init__(
            match_weight=match_weight,
            mismatch_weight=mismatch_weight,
            gap_weight=gap_weight,
            match_dict=match_dict,
            gap_char=gap_char,
        )



    # The auxilary backtrack function.
    def backtrack(self,
        score_matrix: np.ndarray,
        str1: Union[str, List[str]],
        str2: Union[str, List[str]],
    ) -> Tuple[Union[str, List[str]], Union[str, List[str]]]:
        r"""
        This function is an auxilary function, used by the get_alignment() function, that backtracks the score matrix to get the aligned strings.

        Arguments:
            score_matrix (np.ndarray): The score matrix.
            str1: The first string (or list of strings).
            str2: The second string (or list of strings).

        Returns:
            The aligned strings (or list of strings). The aligned strings are padded with spaces to make them the same length.

        .. note::
            * The score matrix is assumed to be a 2D numpy array.
            * There might be multiple optimal alignments. This function returns one of the optimal alignments.
            * The backtracking step has a time complexity of :math:`O(m + n)`, where :math:`n` and :math:`m` are the lengths of the strings str1 and str2, respectively.
        """

        # Lengths of strings str1 and str2, respectively.
        len1 = len(str1)
        len2 = len(str2)

        # Initialize the aligned strings.
        aligned_str1 = ""
        aligned_str2 = ""

        # Initialize the current position.
        i = len1
        j = len2

        # Backtrack until the current position is (0, 0).
        while i > 0 and j > 0:
            # If the current position is the result of a match/mismatch, add the characters to the aligned strings and move to the diagonal.
            if score_matrix[i, j] == score_matrix[i - 1, j - 1] + self.get_score(str1[i - 1], str2[j - 1]):
                insert_str1, insert_str2 = self.add_space_to_shorter(str1[i - 1], str2[j - 1])
                i -= 1
                j -= 1
            # If the current position is the result of a gap in str1, add a gap to str1 and the character to str2 and move to the left.
            elif score_matrix[i, j] == score_matrix[i, j - 1] + self.get_gap_weight(str2[j - 1]):
                insert_str1, insert_str2 = self.add_space_to_shorter(self.gap_char, str2[j - 1])
                j -= 1
            # If the current position is the result of a gap in str2, add a gap to str2 and the character to str1 and move up.
            elif score_matrix[i, j] == score_matrix[i - 1, j] + self.get_gap_weight(str1[i - 1]):
                insert_str1, insert_str2 = self.add_space_to_shorter(str1[i - 1], self.gap_char)
                i -= 1

            # Add the characters to the aligned strings.
            aligned_str1 = insert_str1 + ' | ' + aligned_str1
            aligned_str2 = insert_str2 + ' | ' + aligned_str2

        # If there are still characters in str1, add them to the aligned strings.
        while i > 0:
            insert_str1, insert_str2 = self.add_space_to_shorter(str1[i - 1], self.gap_char)
            aligned_str1 = insert_str1 + ' | ' + aligned_str1
            aligned_str2 = insert_str2 + ' | ' + aligned_str2
            i -= 1

        # If there are still characters in str2, add them to the aligned strings.
        while j > 0:
            insert_str1, insert_str2 = self.add_space_to_shorter(self.gap_char, str2[j - 1])
            aligned_str1 = insert_str1 + ' | ' + aligned_str1
            aligned_str2 = insert_str2 + ' | ' + aligned_str2
            j -= 1

        # Remove the last ' | ' from the aligned strings.
        aligned_str1 = aligned_str1[:-3]
        aligned_str2 = aligned_str2[:-3]

        # Return the aligned strings.
        return aligned_str1, aligned_str2



    # Get the alignment of two strings (or list of strings).
    def get_alignment(self,
        str1: Union[str, List[str]],
        str2: Union[str, List[str]],
        return_score_matrix: bool = False,
    ) -> Tuple[Union[str, List[str]], Union[str, List[str]], Optional[np.ndarray]]:
        r"""
        This is the main function in the NeedlemanWunsch class that gets the alignment of two strings (or list of strings) by using the Needleman-Wunsch algorithm.

        Arguments:
            str1: The first string (or list of strings).
            str2: The second string (or list of strings).
            return_score_matrix (bool): Whether to return the score matrix. (Default: False)

        Returns:
            The aligned strings (or list of strings). The aligned strings are padded with spaces to make them the same length. If return_score_matrix is True, the score matrix is also returned.

        .. note::
            * There might be multiple optimal alignments. This function returns one of the optimal alignments.
            * The time complexity of this function is :math:`O(nm)`, where :math:`n` and :math:`m` are the lengths of the strings str1 and str2, respectively.
            * The space complexity of this function is :math:`O(nm)`.
            * The " | " character is used to separate the elements in the aligned strings.
        """

        # Lengths of strings str1 and str2, respectively.
        len1 = len(str1)
        len2 = len(str2)

        # Initialize the score matrix.
        score_matrix = np.zeros((len1 + 1, len2 + 1))

        # Initialize the first row and column of the score matrix.
        for i in range(1, len1 + 1):
            score_matrix[i, 0] = score_matrix[i - 1, 0] + self.get_gap_weight(str1[i - 1])
        for j in range(1, len2 + 1):
            score_matrix[0, j] = score_matrix[0, j - 1] + self.get_gap_weight(str2[j - 1])

        # Dynamic programming solution (Needleman-Wunsch algorithm):
        for i in range(1, len1 + 1):
            for j in range(1, len2 + 1):
                # Get the scores of the three possible paths.
                match_score = score_matrix[i - 1, j - 1] + self.get_match_weight(str1[i - 1], str2[j - 1])
                delete_score = score_matrix[i - 1, j] + self.get_gap_weight(str1[i - 1])
                insert_score = score_matrix[i, j - 1] + self.get_gap_weight(str2[j - 1])

                # Get the maximum score.
                max_score = max(match_score, delete_score, insert_score)

                # Fill the score matrix.
                score_matrix[i, j] = max_score

        # Get the alignment.
        aligned_str1, aligned_str2 = self.backtrack(score_matrix, str1, str2)

        # Return the alignment and the score matrix.
        if return_score_matrix:
            return aligned_str1, aligned_str2, score_matrix
        return aligned_str1, aligned_str2



# Hirschberg algorithm (linear space algorithm).
class Hirschberg(NeedlemanWunsch):
    def __init__(self,
        match_weight: Union[int, float] = 2,
        mismatch_weight: Union[int, float] = -1,
        gap_weight: Union[int, float] = -2,
        gap_char: str = '-',
        match_dict: dict = None,
    ) -> None:
        r"""
        This function initializes the parameters of the Hirschberg algorithm [H1975]_, a space-efficient solution to the global alignment problem. It inherits from the NeedlemanWunsch class.

        Arguments:
            match_weight (int or float): The weight of a match (default: 2).
            mismatch_weight (int or float): The weight of a mismatch (default: -1).
            gap_weight (int or float): The weight of a gap (default: -2).
            gap_char (str): The character used to represent a gap (default: '-').
            match_dict (dict): The dictionary that maps the characters to their match weights (default: None).

        .. note::
            * The default values are the same as the ones used in the Needleman-Wunsch algorithm.
            * The time complexity of Hirschberg's algorithm is :math:`O(nm)`, where :math:`n` and :math:`m` are the lengths of the strings str1 and str2, respectively.
            * The space complexity of Hirschberg's algorithm is, on the other hand, :math:`O(min(n, m))`.
            * We benefited from the following resources to implement this class: [K2015]_, [W2012]_, [M2010]_, [K2002]_.

        .. [H1975] Hirschberg, Daniel S. "A linear space algorithm for computing maximal common subsequences." Communications of the ACM 18.6 (1975): 341-343.
        .. [K2015] Kellis, Manolis. Computational Biology: Genomes, Networks, Evolution (MIT Course 6.047/6.878) — https://ocw.mit.edu/ans7870/6/6.047/f15/MIT6_047F15_Compiled.pdf (Accessed on 02-16-2023) (Section 2.5.8; Linear Space Alignment, pg. 38-39).
        .. [W2012] Wayne, Kevin. Lecture Slides for Algorithm Design - Dynamic Programming II (https://www.cs.princeton.edu/~wayne/kleinberg-tardos/pdf/06DynamicProgrammingII.pdf) (Accessed on 02-16-2023).
        .. [M2010] Moura, Lucia. Algorithms in Bioinformatics: Lectures 3-5 - Sequence Similarity. Fall 2010. University of Ottawa. (https://www.site.uottawa.ca/~lucia/courses/2010/comp5511/lectures/03-05.pdf) (Accessed on 02-16-2023).
        .. [K2002] Kingsford, Carl. Lecture 7: Dynamic Programming. 2002. Carnegie Mellon University. (https://www.cs.cmu.edu/~ckingsf/class/02-714/Lec07-linspace.pdf) (Accessed on 02-16-2023).
        """

        # Initialize the Needleman-Wunsch algorithm using the super() function.
        super().__init__(
            match_weight=match_weight,
            mismatch_weight=mismatch_weight,
            gap_weight=gap_weight,
            match_dict=match_dict,
            gap_char=gap_char,
        )



    # Get the alignment of two strings (or list of strings).
    def get_alignment(self,
        str1: Union[str, List[str]],
        str2: Union[str, List[str]],
    ) -> Tuple[Union[str, List[str]], Union[str, List[str]]]:
        r"""
        This function gets the alignment of two strings (or list of strings) by using the Hirschberg algorithm.

        Arguments:
            str1: The first string (or list of strings).
            str2: The second string (or list of strings).

        Returns:
            The aligned strings as a tuple of two strings (or list of strings).

        .. note::
            * As a notable improvement of the Needleman-Wunsch algorithm, Hirschberg's algorithm combines both divide-and-conquer and dynamic programming principles. This algorithm provides a space-efficient solution, with a time complexity of :math:`O(nm)`, where :math:`n` and :math:`m` are the lengths of the strings str1 and str2, respectively. Its space complexity, however, is :math:`O(min(n, m))`.
            * To further improve the algorithm, one may limit the number of insertions and deletions in the alignment. This strategy can notably reduce the time complexity from :math:`O(mn)` to :math:`O((m + n) * k)`, where k is the maximum number of insertions or deletions allowed. By constraining the alignment around the diagonal in the score matrix with (2 * k + 1) cells, the new version of the algorithm can be called the k-banded Hirschberg algorithm. If k is arbitrarily small, this modification can lead to a significant improvement in the time complexity.
            * The k-banded Hirschberg algorithm, with its time complexity of :math:`O((m + n) * k)`, is a powerful strategy that balances space and time requirements in sequence alignment.
        """

        # Lengths of strings str1 and str2, respectively.
        len1 = len(str1)
        len2 = len(str2)

        # Check if the length of str1 is less than or equal to the length of str2.
        if len1 >= len2:
            # Get the alignment.
            aligned_str1, aligned_str2 = self.get_alignment_helper(str1, str2)
        else:
            # Get the alignment.
            aligned_str2, aligned_str1 = self.get_alignment_helper(str2, str1)

        # Remove the trailing " | " from the aligned strings (if any).
        aligned_str1 = aligned_str1.strip(" | ")
        aligned_str2 = aligned_str2.strip(" | ")

        # Replace the " |  | " with " | " in the aligned strings (if any).
        aligned_str1 = aligned_str1.replace('|  |', '|')
        aligned_str2 = aligned_str2.replace('|  |', '|')

        # Return the alignment.
        return aligned_str1, aligned_str2



    # Get the alignment of two strings (or list of strings).
    def get_alignment_helper(self,
        str1: Union[str, List[str]],
        str2: Union[str, List[str]],
    ) -> Tuple[Union[str, List[str]], Union[str, List[str]]]:
        """
        This is a helper function that is called by the get_alignment() function. This function gets the alignment of two strings (or list of strings) by using the Hirschberg algorithm.

        Arguments:
            str1: The first string (or list of strings).
            str2: The second string (or list of strings).

        Returns:
            The aligned strings as a tuple of two strings (or list of strings).

        .. note::
            * We assume that the length of str1 is greater than or equal to the length of str2.
        """

        # Lengths of strings str1 and str2, respectively.
        len1 = len(str1)
        len2 = len(str2)

        # Initialize the aligned strings.
        aligned_str1 = ""
        aligned_str2 = ""

        # Check if the length of str1 is 0.
        if len1 == 0:
            # Add gap characters to the shorter string (i.e., str1).
            for j in range(1, len2+1):
                insert_str1, insert_str2 = self.add_space_to_shorter(self.gap_char, str2[j-1])
                aligned_str1 = aligned_str1 + ' | ' + insert_str1
                aligned_str2 = aligned_str2 + ' | ' + insert_str2
        elif len2 == 0:
            # Add gap characters to the shorter string (i.e., str2).
            for i in range(1, len1+1):
                insert_str1, insert_str2 = self.add_space_to_shorter(str1[i-1], self.gap_char)
                aligned_str1 = aligned_str1 + ' | ' + insert_str1
                aligned_str2 = aligned_str2 + ' | ' + insert_str2
        elif len1 == 1 or len2 == 1:
            # Get the alignment of two strings (or list of strings) by using the Needleman-Wunsch algorithm.
            aligned_str1, aligned_str2 = super().get_alignment(str1, str2)
        else:
            # Get the middle index of str1.
            mid1 = len1 // 2

            # Get the scores of the left and right substrings.
            score_row_left = self.nw_score(str1[:mid1], str2)
            # Score-Right = Reverse ( NW-Score( Reverse(Str1-Mid1), Reverse(Str2) ) )
            score_row_right = self.nw_score(str1[mid1:][::-1], str2[::-1])[::-1]

            # Get mid2 = arg max score_row_left + score_row_right
            mid2 = self.get_middle_index(score_row_left, score_row_right)

            # Get the alignment of the left and right substrings.
            aligned_str1_left, aligned_str2_left = self.get_alignment_helper(str1[:mid1], str2[:mid2])
            aligned_str1_right, aligned_str2_right = self.get_alignment_helper(str1[mid1:], str2[mid2:])

            # Combine the aligned strings.
            # Make sure to add ' | ' between the aligned strings only if the aligned strings are not empty.
            # This is to avoid adding ' | ' at the beginning and end of the aligned strings.
            if aligned_str1_left != "" and aligned_str1_right != "":
                aligned_str1 = aligned_str1_left + ' | ' + aligned_str1_right
            else:
                aligned_str1 = aligned_str1_left + aligned_str1_right
            if aligned_str2_left != "" and aligned_str2_right != "":
                aligned_str2 = aligned_str2_left + ' | ' + aligned_str2_right
            else:
                aligned_str2 = aligned_str2_left + aligned_str2_right

        # Return the aligned strings.
        return aligned_str1, aligned_str2



    # Return the last row of the score matrix.
    def nw_score(self,
        str1: Union[str, List[str]],
        str2: Union[str, List[str]],
    ) -> List[float]:
        """
        This function returns the last row of the score matrix.

        Arguments:
            str1: The first string (or list of strings).
            str2: The second string (or list of strings).

        Returns:
            The last row of the score matrix.
        """

        # Lengths of strings str1 and str2, respectively.
        len1 = len(str1)
        len2 = len(str2)

        # Create a 2 x (len2 + 1) matrix.
        score_matrix = np.zeros((2, len2 + 1))

        # Initialize the first row of the score matrix.
        for j in range(1, len2 + 1):
            score_matrix[0, j] = score_matrix[0, j - 1] + self.get_gap_weight(str2[j - 1]) # insertion cost

        # Update the score matrix.
        for i in range(1, len1 + 1):
            score_matrix[1, 0] = score_matrix[0, 0] + self.get_gap_weight(str1[i - 1]) # deletion cost

            for j in range(1, len2 + 1):
                score_matrix[1, j] = max(
                    score_matrix[0, j - 1] + self.get_score(str1[i - 1], str2[j - 1]), # match/mismatch cost
                    score_matrix[0, j] + self.get_gap_weight(str1[i - 1]), # deletion cost
                    score_matrix[1, j - 1] + self.get_gap_weight(str2[j - 1]) # insertion cost
                )

            # Update the score matrix.
            score_matrix[0, :] = score_matrix[1, :]

        # Return the last row of the score matrix.
        return score_matrix[1, :]



    # Get the middle index of str2.
    def get_middle_index(self,
        score_left: List[float],
        score_right: List[float],
    ) -> int:
        """
        This function gets the middle index of str2.

        Arguments:
            score_left: The score of the left sublist.
            score_right: The score of the right sublist.

        Returns:
            The middle index of str2.
        """

        # Length of score_left.
        len_score_left = len(score_left)

        # Initialize the middle index.
        mid2 = 0

        # Initialize the maximum score with the possible minimum score.
        # Oh dear, initially I used 0 as the maximum score, but that was wrong.
        # The maximum score can be negative, so we need to use the possible minimum score instead, which is -float('inf').
        max_score = -float('inf')

        # Get the middle index.
        for i in range(len_score_left):
            if score_left[i] + score_right[i] > max_score:
                mid2 = i
                max_score = score_left[i] + score_right[i]

        # Return the middle index.
        return mid2



# Smith-Waterman algorithm (local alignment).
class SmithWaterman(NeedlemanWunsch):
    def __init__(self,
        match_weight: Union[int, float] = 1,
        mismatch_weight: Union[int, float] = -1,
        gap_weight: Union[int, float] = -1,
        gap_char: str = '-',
        match_dict: dict = None,
    ) -> None:
        r"""
        This function initializes the class variables of the Smith-Waterman algorithm, used for local alignment of sequences (e.g., strings or lists of strings) such as DNA sequences.

        Arguments:
            match_weight (int or float): The weight of a match (default: 1).
            mismatch_weight (int or float): The weight of a mismatch (default: -1).
            gap_weight (int or float): The weight of a gap (default: -1).
            gap_char (str): The character used to represent a gap (default: '-').
            match_dict (dict): The dictionary that maps the characters to their match weights (default: None).

        .. note::
            * The default values are the same as the Needleman-Wunsch algorithm.
            * The Smith-Waterman algorithm can be thought of a variant of the Needleman-Wunsch algorithm, where the max function is replaced by the max function with 0 as the default value and the first row and column of the score matrix are initialized to 0.

            .. math::
                :nowrap:

                \begin{align}
                    \texttt{max score} &= \texttt{max}(\texttt{match score}, \texttt{delete score}, \texttt{insert score}, 0)
                \end{align}

            * We only need to override the get_alignment and backtrack functions of the NeedlemanWunsch class.
            * The space and time complexities of the Smith-Waterman algorithm are the same as the Needleman-Wunsch algorithm, that is, :math:`O(mn)` and :math:`O(mn)`, respectively, where :math:`n` and :math:`m` are the lengths of the two strings (or lists of strings).
        """

        # Initialize the SmithWaterman class using its parent class, NeedlemanWunsch.
        super().__init__(
            match_weight=match_weight,
            mismatch_weight=mismatch_weight,
            gap_weight=gap_weight,
            match_dict=match_dict,
            gap_char=gap_char,
        )



    # Backtrack the score matrix.
    # Override the backtrack function of the NeedlemanWunsch class.
    def backtrack(self,
        score_matrix: np.ndarray,
        str1: Union[str, List[str]],
        str2: Union[str, List[str]],
    ) -> Tuple[Union[str, List[str]], Union[str, List[str]]]:
        """
        This function overrides the backtrack function of the NeedlemanWunsch class to get an optimal local alignment between two strings (or list of strings).

        Arguments:
            score_matrix (numpy.ndarray): The score matrix.
            str1 (str or list of str): The first string (or list of strings).
            str2 (str or list of str): The second string (or list of strings).

        Returns:
            The aligned substrings as a tuple of two strings (or list of strings).

        .. note::
            * The backtrack function used in this function is different from the backtrack function used in the Needleman-Wunsch algorithm. Here we start from the position with the highest score in the score matrix and trace back to the first position that has a score of zero. This is because the highest-scoring subsequence may not necessarily span the entire length of the sequences being aligned.
            * On the other hand, the backtrack function used in the Needleman-Wunsch algorithm traces back through the entire score matrix, starting from the bottom-right corner, to determine the optimal alignment path. This is because the algorithm seeks to find the global alignment of two sequences, which means aligning them from the beginning to the end.
        """

        # Initialize the aligned substrings.
        aligned_str1 = ""
        aligned_str2 = ""

        # Get the position with the maximum score in the score matrix.
        # TODO(msuzgun): See if there is a faster way to get the position with the maximum score in the score matrix.
        i, j = np.unravel_index(np.argmax(score_matrix, axis=None), score_matrix.shape)

        # Backtrack the score matrix.
        while score_matrix[i, j] != 0:
            # Get the scores of the three possible paths.
            match_score = score_matrix[i - 1, j - 1] + self.get_match_weight(str1[i - 1], str2[j - 1])
            delete_score = score_matrix[i - 1, j] + self.get_gap_weight(str1[i - 1])
            insert_score = score_matrix[i, j - 1] + self.get_gap_weight(str2[j - 1])

            # Get the maximum score.
            max_score = max(match_score, delete_score, insert_score)

            # Backtrack the score matrix.
            if max_score == match_score:
                insert_str1, insert_str2 = self.add_space_to_shorter(str1[i - 1], str2[j - 1])
                i -= 1
                j -= 1
            elif max_score == delete_score:
                insert_str1, insert_str2 = self.add_space_to_shorter(str1[i - 1], self.gap_char)
                i -= 1
            elif max_score == insert_score:
                insert_str1, insert_str2 = self.add_space_to_shorter(self.gap_char, str2[j - 1])
                j -= 1

            # Add the characters to the aligned strings.
            aligned_str1 = insert_str1 + ' | ' + aligned_str1
            aligned_str2 = insert_str2 + ' | ' + aligned_str2

        # Remove the last ' | '.
        aligned_str1 = aligned_str1[:-3]
        aligned_str2 = aligned_str2[:-3]

        # Return the aligned substrings.
        return aligned_str1, aligned_str2




    # Get the alignment of two strings (or list of strings).
    # Override the get_alignment function of the NeedlemanWunsch class.
    def get_alignment(self,
        str1: Union[str, List[str]],
        str2: Union[str, List[str]],
        return_score_matrix: bool = False,
    ) -> Tuple[Union[str, List[str]], Union[str, List[str]]]:
        """
        This function overrides the get_alignment function of the NeedlemanWunsch class to get the alignment of two strings (or list of strings) by using the Smith-Waterman algorithm.

        Arguments:
            str1 (str or list of str): The first string (or list of strings).
            str2 (str or list of str): The second string (or list of strings).
            return_score_matrix (bool): Whether to return the score matrix (default: False)

        Returns:
            The aligned strings as a tuple of two strings (or list of strings). If return_score_matrix is True, the score matrix is also returned.

        .. note::
            * The Smith-Waterman algorithm is a dynamic programming algorithm that finds the optimal local alignment between two strings (or list of strings).
            * This function is similar to the get_alignment function in the NeedlemanWunsch class, with two differences. First, the first row and column of the score matrix are initialized to 0. Second, the max function used in the dynamic programming solution is replaced with a max function that defaults to 0, i.e., max_score = max(match_score, delete_score, insert_score, 0).
            * Despite these differences, the time and space complexity of the Smith-Waterman algorithm remain the same as that of the Needleman-Wunsch algorithm. It should be noted that most of the code in this function is identical to that in the get_alignment function of the NeedlemanWunsch class.
        """

         # Lengths of strings str1 and str2, respectively.
        len1 = len(str1)
        len2 = len(str2)

        # Initialize the score matrix.
        score_matrix = np.zeros((len1 + 1, len2 + 1))

        # Initialize the first row and column of the score matrix.
        # This time the first row and column are initialized to 0.

        # Dynamic programming solution (Needleman-Wunsch algorithm):
        for i in range(1, len1 + 1):
            for j in range(1, len2 + 1):
                # Get the scores of the three possible paths.
                match_score = score_matrix[i - 1, j - 1] + self.get_match_weight(str1[i - 1], str2[j - 1])
                delete_score = score_matrix[i - 1, j] + self.get_gap_weight(str1[i - 1])
                insert_score = score_matrix[i, j - 1] + self.get_gap_weight(str2[j - 1])

                # Get the maximum score.
                # Note that this is the only difference between the Smith-Waterman algorithm and the Needleman-Wunsch algorithm.
                # The max function is replaced by the max function with 0 as the default value.
                max_score = max(match_score, delete_score, insert_score, 0.)

                # Fill the score matrix.
                score_matrix[i, j] = max_score

        # Get the alignment.
        aligned_str1, aligned_str2 = self.backtrack(score_matrix, str1, str2)

        # Return the alignment and the score matrix.
        if return_score_matrix:
            return aligned_str1, aligned_str2, score_matrix
        return aligned_str1, aligned_str2



# Dynamic time warping (DTW) class.
class DTW:
    def __init__(self) -> None:
        r"""
        This function initializes the Dynamic time warping (DTW) class.
        """
        pass


    # Get the alignment indices of two sequences (or list of sequences) by using the DTW algorithm.
    def get_alignment_path(self,
        sequence1: Union[str, List[str], int, List[int], float, List[float], np.ndarray],
        sequence2: Union[str, List[str], int, List[int], float, List[float], np.ndarray],
        distance = 'absolute_difference',
        p_value: Optional[int] = None,
    ) -> List[Tuple[int, int]]:
        """
        This function gets the alignment indices of two sequences (or list of sequences) by using the DTW algorithm.

        Arguments:
            sequence1: The first sequence.
            sequence2: The second sequence.
            distance (str): The distance function to be used (currently only 'absolute_difference' and 'square_difference' are supported).

        Returns:
            The path of the alignment as a list of tuples of two integers.

        Raises:
            TypeError: If the input sequences are not of the same type.
            ValueError: If the distance function is not supported.


        .. note::
            * The DTW algorithm is a dynamic programming algorithm that finds the optimal alignment between two sequences (or list of sequences).
            * The time complexity of the DTW algorithm is :math:`O(nm)`, where :math:`n` and :math:`m` are the lengths of the two sequences, respectively.
        """

        # First check if both sequences are of the same type.
        if type(sequence1) != type(sequence2):
            raise TypeError("Both sequences must be of the same type.")

        # Check if the distance function is supported.
        if distance not in ['absolute_difference', 'square_difference']:
            raise ValueError("The distance function must be either 'absolute_difference' or 'square_difference'.")

        # If the sequences are strings or lists of strings, convert them to lists of integers (ASCII codes), in np.ndarray format.
        if type(sequence1) == str or (type(sequence1) == list and type(sequence1[0]) == str):
            sequence1 = np.array([ord(char) for char in sequence1])
            sequence2 = np.array([ord(char) for char in sequence2])

        # Get the lengths of the sequences.
        len1 = len(sequence1)
        len2 = len(sequence2)

        # Initialize the DTW distance matrix with infinity values.
        distance_matrix = np.full((len1 + 1, len2 + 1), np.inf)

        # Initialize the first row and column of the DTW distance matrix with zero.
        distance_matrix[0, 0] = 0.

        # Fill the DTW distance matrix.
        for i in range(1, len1 + 1):
            for j in range(1, len2 + 1):
                # Get the distance between the two elements.
                if distance == 'absolute_difference':
                    distance = abs(sequence1[i - 1] - sequence2[j - 1])
                else:
                    # distance == 'square_difference'
                    distance = (sequence1[i - 1] - sequence2[j - 1]) ** 2

                # Fill the DTW distance matrix.
                distance_matrix[i, j] = distance + min(
                        distance_matrix[i - 1, j],
                        distance_matrix[i, j - 1],
                        distance_matrix[i - 1, j - 1]
                    )

        # Initialize the alignment.
        alignment = []

        # Get the alignment.
        i = len1
        j = len2
        while i > 0 or j > 0:
            alignment.append((i - 1, j - 1))
            if i == 0:
                j -= 1
            elif j == 0:
                i -= 1
            else:
                if distance_matrix[i - 1, j] < distance_matrix[i, j - 1] and distance_matrix[i - 1, j] < distance_matrix[i - 1, j - 1]:
                    i -= 1
                elif distance_matrix[i, j - 1] < distance_matrix[i - 1, j - 1]:
                    j -= 1
                else:
                    i -= 1
                    j -= 1

        # Reverse the alignment.
        alignment.reverse()

        # Return the alignment.
        return alignment

# Longest common subsequence (LCSubsequenceuence) class
class LongestCommonSubsequence(StringAlignment):
    # Initialize the class
    def __init__(self,
        list_of_list_separator: str = " ## ",
        ) -> None:
        r"""
        This function initializes the Longest Common Subsequence (LCSubsequenceuence) class, which inherits from the StringAlignment class.

        Longest common subsequence (LCSubsequence) of two strings is a subsequence of maximal length that appears in both of them.

        The following recurrence relation can be used to solve the LCSubsequence problem:

        .. math::
            :nowrap:

            \begin{align}
                L[i,j] =
                \begin{cases}
                    0 &\text{ if } i=0 \text{ or } j=0\\
                    L[i-1,j-1]+1 &\text{ if } i,j>0 \text{ and } str1[i]=str2[j]\\
                    \max(L[i-1,j],L[i,j-1]) &\text{ if } i,j>0 \text{ and } str1[i]\neq str2[j]\\
                \end{cases}
            \end{align}

        where :math:`L[i,j]` denotes the length of the LCSubsequence of the prefixes str1[0:i] and str2[0:j]. The solution to the problem is then given by :math:`L[n,m]`, assuming that str1 and str2 have lengths n and m, respectively.

        A dynamic programming solution exists for this problem with a quadratic (i.e., :math:`\mathcal{O}(nm)`) space and time complexity.

        If the vocabulary is fixed, LCSubsequence admits a "Four-Russians speedup," which reduces its overall time complexity to subquadratic :math:`\mathcal{O}(n^2/\log n)`, but this algorithm is not yet implemented in this package.

        Arguments:
            list_of_list_separator (str): Separator to use when the inputs are lists of strings.
        """
        # Initialize the StringAlignment class.
        super().__init__(match_weight=1, mismatch_weight=0, gap_weight=0)

        # Set the list of list separator.
        self.list_of_list_separator = list_of_list_separator



    # Compute the longest common subsequence between two strings
    def compute(self,
        str1: Union[str, List[str]],
        str2: Union[str, List[str]],
        returnCandidates: bool = False,
    ) -> Tuple[float, Union[List[str], List[List[str]]]]:
        """
        This function computes the longest common subsequence between two strings (or lists of strings).

        Arguments:
            str1 (str or list of str): The first string (or list of strings) to compare.
            str2 (str or list of str): The second string (or list of strings).
            returnCandidates (bool): Whether to return the candidates for the longest common subsequence (default: False).

        Returns:
            * If returnCandidates is False, then the length of the longest common subsequence between the two strings.
            * If returnCandidates is True, then the set of all candidates for the longest common subsequence is also returned.

        .. note::
           * Similar to that of the Levenshtein edit distance problem, the dynamic programming solution for the longest common subsequence problem can be further optimized by using the last row of the two-dimensional array L to compute the length of the LCSubsequence. The key idea behind this optimization is that the last row of the array L only depends on the values of the previous row. Therefore, we can store only two rows of the array L at a time and compute the LCSubsequence using these two rows.
           * This optimization reduces the space complexity of the algorithm to :math:`\mathcal{O}(m)`, where :math:`m` is the length of the shorter input string. This optimization is particularly useful when one of the input strings is much shorter than the other, as it can significantly reduce the amount of memory required to solve the problem.
        """
        # Check whether the inputs are lists of strings.
        boolList = False
        if isinstance(str1, list) and isinstance(str2, list):
            boolList = True

        # Lengths of strings str1 and str2, respectively.
        n = len(str1)
        m = len(str2)

        # Initialize the distance matrix.
        dist = np.zeros((n + 1, m + 1))

        # Dynamic programming solution to the longest common subsequence.
        for i in range(1, n + 1):
            for j in range(1, m + 1):
                # if str1[i-1] == str2[j-1]: # This is the original code. changed: 2023-03-19, 10:05 PM
                if self.bool_match(str1[i-1], str2[j-1]):
                    dist[i, j] = dist[i-1, j-1] + 1
                else:
                    dist[i, j] = max(dist[i-1, j], dist[i, j-1])

        # TODO(msuzgun): At the moment, the backtrack function is not optimized and pretty slow. It should be optimized!
        def backtrack(i: int, j: int) -> Union[List[str], List[List[str]]]:
            """
            This function, which is called recursively and inside the compute function, is used to backtrack the distance matrix to compute the longest common subsequence.

            Arguments:
                i (int): The row index of the distance matrix.
                j (int): The column index of the distance matrix.

            Returns:
                The set of longest common subsequences between the two strings (or lists of strings).
            """
            # If the row or column index is 0, then the longest common subsequence is empty.
            if i == 0 or j == 0:
                # return [''] if boolList else [] // This is the original code.
                return []

            # If the characters at the current row and column of the distance matrix are equal, then the current character is part of the longest common subsequence.
            # if str1[i-1] == str2[j-1]: # This is the original code. changed: 2023-03-19, 10:05 PM
            if self.bool_match(str1[i-1], str2[j-1]):
                # insert_elt = str1[i-1] if boolList else str1[i-1] // This is the original code.
                insert_elt = [str1[i-1]] if boolList else str1[i-1]
                candidates = list(
                    set(
                        cartesian_product(
                            backtrack(i-1, j-1),
                            insert_elt,
                            boolList=boolList,
                            list_of_list_separator=self.list_of_list_separator,
                        )
                    )
                )
                return candidates

            # If the characters at the current row and column of the distance matrix are not equal, then the current character is not part of the longest common subsequence.
            candidates = []
            if dist[i, j-1] >= dist[i-1, j]:
                candidates = backtrack(i, j-1)
            if dist[i-1, j] >= dist[i, j-1]:
                candidates += backtrack(i-1, j)
            return list(set(candidates))

        # Compute the longest common subsequence.
        candidates = None
        if returnCandidates:
            candidates = backtrack(n, m)
            if boolList:
                candidates = [
                    elt.split(self.list_of_list_separator) for elt in candidates
                ]
        return dist[n, m], candidates




# Longest common substring (LCSubstring) class
class LongestCommonSubstring(LongestCommonSubsequence):
    # Initialize the class
    def __init__(self,
        list_of_list_separator: str = " ## ",
        ) -> None:
        r"""
        This function initializes the LongestCommonSubstring (LCSubstring) class.

        Longest Common Substring (LCSubstring) of two strings is the longest substring that appears in both of them.

        The following recurrence relation can be used to solve the LCSubstring problem:

        .. math::
            :nowrap:

            \begin{align}
                L[i,j] =
                \begin{cases}
                    0 &\text{ if } i=0 \text{ or } j=0\\
                    L[i-1,j-1]+1 &\text{ if } i,j>0 \text{ and } str1[i]=str2[j]\\
                    0 &\text{ if } i,j>0 \text{ and } str1[i]\neq str2[j]\\
                \end{cases}
            \end{align}

        where :math:`L[i,j]` denotes the length of the LCSubstring that ends at indices i and j in str1 and str2, respectively. The solution to the problem is then given by the maximum value of :math:`L[i,j]`, assuming that str1 and str2 have lengths n and m, respectively.

        A dynamic programming solution exists for this problem with a quadratic (i.e., :math:`\mathcal{O}(nm)`) space and time complexity.

        Arguments:
            list_of_list_separator (str): Separator to use when the inputs are lists of strings.

        Returns:
            None
        """
        # Separator to use when the inputs are lists of strings.
        super().__init__(list_of_list_separator=list_of_list_separator)



    # Compute the longest common substring between two strings
    def compute(self,
        str1: Union[str, List[str]],
        str2: Union[str, List[str]],
        returnCandidates: bool = False,
    ) -> Tuple[float, Union[List[str], List[List[str]]]]:
        """
        This function computes the longest common substring between two strings (or lists of strings).

        Arguments:
            str1 (str or list of str): The first string (or list of strings).
            str2 (str or list of str): The second string (or list of strings).
            returnCandidates (bool): A boolean flag indicating whether to return the longest common substring as a list of lists.

        Returns:
            If returnCandidates is False, then the length of the longest common substring between the two strings. If returnCandidates is True, then the set of longest common substrings between the two strings (or lists of strings) is also returned.


        .. note::
            * There exists a linear-time solution to LCSubstring problem that uses generalized suffix trees.
            * As with the longest common subsequence problem, the longest common substring is not unique. It is possible to have multiple substrings with the same maximum length that appear in both strings.
            * Similar to the dynamic programming solution for the longest common subsequence problem, the last row of the matrix can also be used to optimize the computation of the longest common substring.
            * It's important to note that the longest common substring is different from the longest common subsequence. The longest common substring is a contiguous sequence of symbols that appears in both strings, while the longest common subsequence is a sequence of characters that may not be contiguous.
            * The longest common substring is a measure of similarity between two strings and is used in various fields, including computational biology, where it is used to compare DNA sequences. Similarly, it is also used in plagirism detection and other applications.
        """
        # Determine whether the inputs are lists of strings.
        boolList = False
        if isinstance(str1, list) and isinstance(str2, list):
            boolList = True

        # Lengths of strings str1 and str2, respectively.
        n = len(str1)
        m = len(str2)

        # Initialize the distance matrix.
        dist = np.zeros((n + 1, m + 1), dtype=int)

        # Initialize the longest common substring length.
        longest_common_substring_length = 0

        # Initialize the longest common substring candidates.
        longest_common_substring_indices = []

        # Dynamic programming solution to the longest common substring.
        for i in range(1, n + 1):
            for j in range(1, m + 1):
                # if str1[i-1] == str2[j-1]: # # This is the original code. changed: 2023-03-19, 10:05 PM
                if self.bool_match(str1[i-1], str2[j-1]):
                    dist[i, j] = dist[i-1, j-1] + 1
                    if dist[i, j] > longest_common_substring_length:
                        longest_common_substring_length = dist[i, j]
                        longest_common_substring_indices = [i]
                        # candidates = [str1[i-longest_common_substring_length:i]]
                    elif dist[i, j] == longest_common_substring_length:
                        # candidates.append(str1[i-longest_common_substring_length:i])
                        longest_common_substring_indices.append(i)
                else:
                    dist[i, j] = 0

        # If returnCandidates is True, then additionally return the set of longest common substrings.
        if returnCandidates:
            longest_common_substring_candidates = [str1[i-longest_common_substring_length:i] for i in longest_common_substring_indices]
            if boolList:
                # TODO(msuzgun): Double check this. Correct, but there might be a better way to do this.
                longest_common_substring_candidates = list(set(
                    [
                        f"{self.list_of_list_separator}".join(cand) for cand in longest_common_substring_candidates
                    ]
                ))
                longest_common_substring_candidates = [
                    cand.split(self.list_of_list_separator) for cand in longest_common_substring_candidates
                ]
                longest_common_substring_candidates = set(tuple(elt) for elt in longest_common_substring_candidates)
            else:
                longest_common_substring_candidates = list(set(longest_common_substring_candidates))
            return longest_common_substring_length, longest_common_substring_candidates
        return longest_common_substring_length, None

"""
string2string code ends here
"""
